"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class GroupManagement1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'GroupManagement1' service.
    """

    SERVICE_MANUFACTURER = 'SonosInc'
    SERVICE_TYPE = 'urn:schemas-upnp-org:service:GroupManagement:1'

    SERVICE_DEFAULT_VARIABLES = {
        "SourceAreaIds": { "data_type": "string", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {
        "GroupCoordinatorIsLocal": { "data_type": "boolean", "default": None, "allowed_list": None},
        "LocalGroupUUID": { "data_type": "string", "default": None, "allowed_list": None},
        "ResetVolumeAfter": { "data_type": "boolean", "default": None, "allowed_list": None},
        "VirtualLineInGroupID": { "data_type": "string", "default": None, "allowed_list": None},
        "VolumeAVTransportURI": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_AddMember(self, MemberID, BootSeq, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the AddMember action.

            :returns: "CurrentTransportSettings", "CurrentURI", "GroupUUIDJoined", "ResetVolumeAfter", "VolumeAVTransportURI"
        """
        arguments = {
            "MemberID": MemberID,
            "BootSeq": BootSeq,
        }

        out_params = self.call_action("AddMember", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentTransportSettings", "CurrentURI", "GroupUUIDJoined", "ResetVolumeAfter", "VolumeAVTransportURI",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_RemoveMember(self, MemberID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RemoveMember action.
        """
        arguments = {
            "MemberID": MemberID,
        }

        self.call_action("RemoveMember", arguments=arguments, aspects=aspects)

        return

    def action_ReportTrackBufferingResult(self, MemberID, ResultCode, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ReportTrackBufferingResult action.
        """
        arguments = {
            "MemberID": MemberID,
            "ResultCode": ResultCode,
        }

        self.call_action("ReportTrackBufferingResult", arguments=arguments, aspects=aspects)

        return

    def action_SetSourceAreaIds(self, DesiredSourceAreaIds, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetSourceAreaIds action.
        """
        arguments = {
            "DesiredSourceAreaIds": DesiredSourceAreaIds,
        }

        self.call_action("SetSourceAreaIds", arguments=arguments, aspects=aspects)

        return



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import NamedTuple

class FastPingResult(NamedTuple):
    pings_sent: int
    pings_received: int
    ping_loss_rate: int



"""
.. module:: entrypoints
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: A set of standaridized entry point functions that provide standardized task orchestration
               of workpacket jobs.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import os
import sys

from akit.xlogging.foundations import logging_initialize, getAutomatonKitLogger
from akit.workflow.execution import execute_workflow

logger = getAutomatonKitLogger()

def run_workflow_entrypoint(workflow_file: str, workflow_info: dict):
    """
        This is the entry point for the execution of workpackets.  It provides a the execution
        context for the execution of the orchestration steps in the workpacket.

        environment:
            AKIT_BUILD_BRANCH: somebranch
            AKIT_BUILD_NAME: somebuild-2.1.456
            AKIT_JOBTYPE: unknown

        parameters:
            branch: somebranch
            build: somebuild-2.1.456
            landscape: $HOME/akit/config/landscape.yaml

        tasklist:
            - label: Print OS Name
              tasktype: akit.workflow.tasks.embeddedpython@EmbeddedPython
              script:
                  - import os
                  - print(os.name)

            - label: List Directories
              tasktype: akit.workflow.tasks.shellscript@BashScript
              script:
                  - ls -al

    """
    # We must exit with a result code, initialize it to 0 here
    result_code = 0

    if "environment" not in workflow_info:
        error_msg = "The work packet file must have an 'workpacket->environment section. file=%s" % workpacket_file
        raise SyntaxError(error_msg)

    if "parameters" not in workflow_info:
        error_msg = "The work packet file must have an 'workpacket->parameters section. file=%s" % workpacket_file
        raise SyntaxError(error_msg)

    if "workflow" not in workflow_info:
        error_msg = "The work packet file must have an 'workpacket->workflow section. file=%s" % workpacket_file
        raise SyntaxError(error_msg)

    environment = workflow_info["environment"]
    del workflow_info["environment"]

    parameters = workflow_info["parameters"]
    del workflow_info["parameters"]

    workflow = workflow_info["workflow"]
    del workflow_info["workflow"]

    execute_workflow(logger, environment=environment, parameters=parameters, workflow=workflow, **workflow_info)

    sys.exit(result_code)

    return



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class VirtualLineIn1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'VirtualLineIn1' service.
    """

    SERVICE_MANUFACTURER = 'SonosInc'
    SERVICE_TYPE = 'urn:schemas-upnp-org:service:VirtualLineIn:1'

    SERVICE_DEFAULT_VARIABLES = {
        "AVTransportURIMetaData": { "data_type": "string", "default": None, "allowed_list": None},
        "CurrentTrackMetaData": { "data_type": "string", "default": None, "allowed_list": None},
        "CurrentTransportActions": { "data_type": "string", "default": None, "allowed_list": None},
        "EnqueuedTransportURIMetaData": { "data_type": "string", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {
        "LastChange": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_Next(self, InstanceID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Next action.
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        self.call_action("Next", arguments=arguments, aspects=aspects)

        return

    def action_Pause(self, InstanceID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Pause action.
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        self.call_action("Pause", arguments=arguments, aspects=aspects)

        return

    def action_Play(self, InstanceID, Speed, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Play action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "Speed": Speed,
        }

        self.call_action("Play", arguments=arguments, aspects=aspects)

        return

    def action_Previous(self, InstanceID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Previous action.
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        self.call_action("Previous", arguments=arguments, aspects=aspects)

        return

    def action_SetVolume(self, InstanceID, DesiredVolume, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetVolume action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "DesiredVolume": DesiredVolume,
        }

        self.call_action("SetVolume", arguments=arguments, aspects=aspects)

        return

    def action_StartTransmission(self, InstanceID, CoordinatorID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the StartTransmission action.

            :returns: "CurrentTransportSettings"
        """
        arguments = {
            "InstanceID": InstanceID,
            "CoordinatorID": CoordinatorID,
        }

        out_params = self.call_action("StartTransmission", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentTransportSettings",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_Stop(self, InstanceID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Stop action.
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        self.call_action("Stop", arguments=arguments, aspects=aspects)

        return

    def action_StopTransmission(self, InstanceID, CoordinatorID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the StopTransmission action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "CoordinatorID": CoordinatorID,
        }

        self.call_action("StopTransmission", arguments=arguments, aspects=aspects)

        return



"""
.. module:: wirelesscoordinatorintegration
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Contains a UpnpCoordinatorIntegration object to use for working with the nodes of a cluster

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Dict, List, Tuple, TYPE_CHECKING

from akit.exceptions import AKitConfigurationError, AKitSemanticError
from akit.coupling.coordinatorcoupling import CoordinatorCoupling
from akit.interop.coordinators.wirelessapcoordinator import WirelessAPCoordinator

# Types imported only for type checking purposes
if TYPE_CHECKING:
    from akit.interop.landscaping.landscape import Landscape

class WirelessCoordinatorIntegration(CoordinatorCoupling):
    """
        The WirelessCoordinatorIntegration handle the requirement registration for the OpenWRT
        wireless router coordinator.
    """

    pathbase = "/wireless"

    def __init__(self, *args, **kwargs):
        """
            The default contructor for an :class:`WirelessCoordinatorIntegration`.
        """
        super(WirelessCoordinatorIntegration, self).__init__(*args, **kwargs)
        return

    @classmethod
    def attach_to_environment(cls, constraints: Dict={}):
        """
            This API is called so that the IntegrationCoupling can process configuration information.  The :class:`IntegrationCoupling`
            will verify that it has a valid environment and configuration to run in.

            :raises :class:`akit.exceptions.AKitMissingConfigError`, :class:`akit.exceptions.AKitInvalidConfigError`:
        """
        resources_acquired = False

        return

    @classmethod
    def attach_to_framework(cls, landscape: "Landscape"):
        """
            This API is called so that the IntegrationCoupling can attach to the test framework and participate with
            registration processes.  This allows the framework to ignore the bringing-up of couplings that are not being
            included by a test.
        """
        super(WirelessCoordinatorIntegration, cls).attach_to_framework(landscape)
        cls.landscape.register_integration_point("coordinator/wireless", cls)
        return

    @classmethod
    def collect_resources(cls):
        """
            This API is called so the `IntegrationCoupling` can connect with a resource management
            system and gain access to the resources required for the automation run.

            :raises :class:`akit.exceptions.AKitResourceError`:
        """
        return

    @classmethod
    def create_coordinator(cls, landscape: "Landscape") -> object:
        """
            This API is called so that the landscape can create a coordinator for a given integration role.
        """
        cls.coordinator = WirelessAPCoordinator(landscape)
        return cls.coordinator

    @classmethod
    def declare_precedence(cls) -> int:
        """
            This API is called so that the IntegrationCoupling can declare an ordinal precedence that should be
            utilized for bringing up its integration state.
        """
        # We need to call the base class, it sets the 'logger' member
        super(WirelessCoordinatorIntegration, cls).declare_precedence()
        return

    @classmethod
    def diagnostic(cls, label: str, level: int, diag_folder: str):
        """
            The API is called by the :class:`akit.sequencer.Sequencer` object when the automation sequencer is
            building out a diagnostic package at a diagnostic point in the automation sequence.  Example diagnostic
            points are:

            * pre-run
            * post-run

            Each diagnostic package has its own storage location so derived :class:`akit.scope.ScopeCoupling` objects
            can simply write to their specified output folder.

            :param label: The label associated with this diagnostic.
            :param level: The maximum diagnostic level to run dianostics for.
            :param diag_folder: The output folder path where the diagnostic information should be written.
        """

        return

    @classmethod
    def establish_connectivity(cls, allow_missing_devices: bool=False, upnp_recording: bool = False, allow_unknown_devices: bool = False) -> Tuple[List[str], dict]:
        """
            This API is called so the `IntegrationCoupling` can establish connectivity with any compute or storage
            resources.

            :returns: A tuple with a list of error messages for failed connections and dict of connectivity
                      reports for devices devices based on the coordinator.
        """

        conn_results = {}

        conn_errors = []

        return (conn_errors, conn_results)

    @classmethod
    def establish_presence(cls) -> Tuple[List[str], dict]:
        """
            This API is called so the `IntegrationCoupling` can establish presence with any compute or storage
            resources.

            :returns: A tuple with a list of error messages for failed connections and dict of connectivity
                      reports for devices devices based on the coordinator.
        """
        cls.coordinator.establish_presence()
        return



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class WFAWLANConfig1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'WFAWLANConfig1' service.
    """

    SERVICE_MANUFACTURER = 'NETGEARInc'
    SERVICE_TYPE = 'urn:schemas-wifialliance-org:service:WFAWLANConfig:1'

    SERVICE_EVENT_VARIABLES = {
        "APStatus": { "data_type": "ui1", "default": None, "allowed_list": None},
        "STAStatus": { "data_type": "ui1", "default": None, "allowed_list": None},
        "WLANEvent": { "data_type": "bin.base64", "default": None, "allowed_list": None},
    }

    def action_DelAPSettings(self, NewAPSettings, extract_returns=True):
        """
            Calls the DelAPSettings action.

            :returns: "result"
        """
        arguments = {
            "NewAPSettings": NewAPSettings,
        }

        out_params = self._proxy_call_action("DelAPSettings", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("result",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_DelSTASettings(self, NewSTASettings, extract_returns=True):
        """
            Calls the DelSTASettings action.

            :returns: "result"
        """
        arguments = {
            "NewSTASettings": NewSTASettings,
        }

        out_params = self._proxy_call_action("DelSTASettings", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("result",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetAPSettings(self, NewMessage, extract_returns=True):
        """
            Calls the GetAPSettings action.

            :returns: "NewAPSettings"
        """
        arguments = {
            "NewMessage": NewMessage,
        }

        out_params = self._proxy_call_action("GetAPSettings", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewAPSettings",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetDeviceInfo(self, extract_returns=True):
        """
            Calls the GetDeviceInfo action.

            :returns: "NewDeviceInfo"
        """
        arguments = { }

        out_params = self._proxy_call_action("GetDeviceInfo", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewDeviceInfo",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetSTASettings(self, NewMessage, extract_returns=True):
        """
            Calls the GetSTASettings action.

            :returns: "NewSTASettings"
        """
        arguments = {
            "NewMessage": NewMessage,
        }

        out_params = self._proxy_call_action("GetSTASettings", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewSTASettings",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_PutMessage(self, NewInMessage, extract_returns=True):
        """
            Calls the PutMessage action.

            :returns: "NewOutMessage"
        """
        arguments = {
            "NewInMessage": NewInMessage,
        }

        out_params = self._proxy_call_action("PutMessage", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewOutMessage",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_PutWLANResponse(self, NewMessage, NewWLANEventType, NewWLANEventMAC, extract_returns=True):
        """
            Calls the PutWLANResponse action.

            :returns: "result"
        """
        arguments = {
            "NewMessage": NewMessage,
            "NewWLANEventType": NewWLANEventType,
            "NewWLANEventMAC": NewWLANEventMAC,
        }

        out_params = self._proxy_call_action("PutWLANResponse", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("result",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_RebootAP(self, NewAPSettings, extract_returns=True):
        """
            Calls the RebootAP action.

            :returns: "result"
        """
        arguments = {
            "NewAPSettings": NewAPSettings,
        }

        out_params = self._proxy_call_action("RebootAP", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("result",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_RebootSTA(self, NewSTASettings, extract_returns=True):
        """
            Calls the RebootSTA action.

            :returns: "result"
        """
        arguments = {
            "NewSTASettings": NewSTASettings,
        }

        out_params = self._proxy_call_action("RebootSTA", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("result",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_ResetAP(self, NewMessage, extract_returns=True):
        """
            Calls the ResetAP action.

            :returns: "result"
        """
        arguments = {
            "NewMessage": NewMessage,
        }

        out_params = self._proxy_call_action("ResetAP", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("result",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_ResetSTA(self, NewMessage, extract_returns=True):
        """
            Calls the ResetSTA action.

            :returns: "result"
        """
        arguments = {
            "NewMessage": NewMessage,
        }

        out_params = self._proxy_call_action("ResetSTA", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("result",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetAPSettings(self, NewAPSettings, extract_returns=True):
        """
            Calls the SetAPSettings action.

            :returns: "result"
        """
        arguments = {
            "NewAPSettings": NewAPSettings,
        }

        out_params = self._proxy_call_action("SetAPSettings", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("result",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetSTASettings(self, extract_returns=True):
        """
            Calls the SetSTASettings action.

            :returns: "NewSTASettings"
        """
        arguments = { }

        out_params = self._proxy_call_action("SetSTASettings", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewSTASettings",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetSelectedRegistrar(self, NewMessage, extract_returns=True):
        """
            Calls the SetSelectedRegistrar action.

            :returns: "result"
        """
        arguments = {
            "NewMessage": NewMessage,
        }

        out_params = self._proxy_call_action("SetSelectedRegistrar", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("result",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class dial1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'dial1' service.
    """

    SERVICE_MANUFACTURER = 'MicrosoftCorporation'
    SERVICE_TYPE = 'urn:dial-multiscreen-org:service:dial:1'

    SERVICE_EVENT_VARIABLES = {}



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class WANIPConnection2ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:WANIPConnection:2' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'WANIPConnection2'

    SERVICE_DEFAULT_VARIABLES = {
        "AutoDisconnectTime": { "data_type": "ui4", "default": None, "allowed_list": None},
        "ConnectionStatus": { "data_type": "string", "default": None, "allowed_list": "['Unconfigured', 'Connecting', 'Connected', 'PendingDisconnect', 'Disconnecting', 'Disconnected']"},
        "ConnectionType": { "data_type": "string", "default": None, "allowed_list": None},
        "ExternalIPAddress": { "data_type": "string", "default": None, "allowed_list": None},
        "ExternalPort": { "data_type": "ui2", "default": None, "allowed_list": None},
        "IdleDisconnectTime": { "data_type": "ui4", "default": None, "allowed_list": None},
        "InternalClient": { "data_type": "string", "default": None, "allowed_list": None},
        "InternalPort": { "data_type": "ui2", "default": None, "allowed_list": None},
        "LastConnectionError": { "data_type": "string", "default": None, "allowed_list": "['ERROR_NONE', 'ERROR_COMMAND_ABORTED', 'ERROR_NOT_ENABLED_FOR_INTERNET', 'ERROR_USER_DISCONNECT', 'ERROR_ISP_DISCONNECT', 'ERROR_IDLE_DISCONNECT', 'ERROR_FORCED_DISCONNECT', 'ERROR_NO_CARRIER', 'ERROR_IP_CONFIGURATION', 'ERROR_UNKNOWN']"},
        "NATEnabled": { "data_type": "boolean", "default": None, "allowed_list": None},
        "PortMappingDescription": { "data_type": "string", "default": None, "allowed_list": None},
        "PortMappingEnabled": { "data_type": "boolean", "default": None, "allowed_list": None},
        "PortMappingLeaseDuration": { "data_type": "ui4", "default": None, "allowed_list": None},
        "PortMappingNumberOfEntries": { "data_type": "ui2", "default": None, "allowed_list": None},
        "PortMappingProtocol": { "data_type": "string", "default": None, "allowed_list": "['TCP', 'UDP']"},
        "PossibleConnectionTypes": { "data_type": "string", "default": None, "allowed_list": None},
        "RSIPAvailable": { "data_type": "boolean", "default": None, "allowed_list": None},
        "RemoteHost": { "data_type": "string", "default": None, "allowed_list": None},
        "SystemUpdateID": { "data_type": "ui4", "default": None, "allowed_list": None},
        "Uptime": { "data_type": "ui4", "default": None, "allowed_list": None},
        "WarnDisconnectDelay": { "data_type": "ui4", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {}

    def action_AddAnyPortMapping(self, NewRemoteHost, NewExternalPort, NewProtocol, NewInternalPort, NewInternalClient, NewEnabled, NewPortMappingDescription, NewLeaseDuration, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the AddAnyPortMapping action.

            :returns: "NewReservedPort"
        """
        arguments = {
            "NewRemoteHost": NewRemoteHost,
            "NewExternalPort": NewExternalPort,
            "NewProtocol": NewProtocol,
            "NewInternalPort": NewInternalPort,
            "NewInternalClient": NewInternalClient,
            "NewEnabled": NewEnabled,
            "NewPortMappingDescription": NewPortMappingDescription,
            "NewLeaseDuration": NewLeaseDuration,
        }

        out_params = self.call_action("AddAnyPortMapping", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewReservedPort",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_AddPortMapping(self, NewRemoteHost, NewExternalPort, NewProtocol, NewInternalPort, NewInternalClient, NewEnabled, NewPortMappingDescription, NewLeaseDuration, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the AddPortMapping action.
        """
        arguments = {
            "NewRemoteHost": NewRemoteHost,
            "NewExternalPort": NewExternalPort,
            "NewProtocol": NewProtocol,
            "NewInternalPort": NewInternalPort,
            "NewInternalClient": NewInternalClient,
            "NewEnabled": NewEnabled,
            "NewPortMappingDescription": NewPortMappingDescription,
            "NewLeaseDuration": NewLeaseDuration,
        }

        self.call_action("AddPortMapping", arguments=arguments, aspects=aspects)

        return

    def action_DeletePortMapping(self, NewRemoteHost, NewExternalPort, NewProtocol, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the DeletePortMapping action.
        """
        arguments = {
            "NewRemoteHost": NewRemoteHost,
            "NewExternalPort": NewExternalPort,
            "NewProtocol": NewProtocol,
        }

        self.call_action("DeletePortMapping", arguments=arguments, aspects=aspects)

        return

    def action_DeletePortMappingRange(self, NewStartPort, NewEndPort, NewProtocol, NewManage, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the DeletePortMappingRange action.
        """
        arguments = {
            "NewStartPort": NewStartPort,
            "NewEndPort": NewEndPort,
            "NewProtocol": NewProtocol,
            "NewManage": NewManage,
        }

        self.call_action("DeletePortMappingRange", arguments=arguments, aspects=aspects)

        return

    def action_ForceTermination(self, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ForceTermination action.
        """
        arguments = { }

        self.call_action("ForceTermination", arguments=arguments, aspects=aspects)

        return

    def action_GetAutoDisconnectTime(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetAutoDisconnectTime action.

            :returns: "NewAutoDisconnectTime"
        """
        arguments = { }

        out_params = self.call_action("GetAutoDisconnectTime", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewAutoDisconnectTime",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetConnectionTypeInfo(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetConnectionTypeInfo action.

            :returns: "NewConnectionType", "NewPossibleConnectionTypes"
        """
        arguments = { }

        out_params = self.call_action("GetConnectionTypeInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewConnectionType", "NewPossibleConnectionTypes",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetExternalIPAddress(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetExternalIPAddress action.

            :returns: "NewExternalIPAddress"
        """
        arguments = { }

        out_params = self.call_action("GetExternalIPAddress", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewExternalIPAddress",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetGenericPortMappingEntry(self, NewPortMappingIndex, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetGenericPortMappingEntry action.

            :returns: "NewRemoteHost", "NewExternalPort", "NewProtocol", "NewInternalPort", "NewInternalClient", "NewEnabled", "NewPortMappingDescription", "NewLeaseDuration"
        """
        arguments = {
            "NewPortMappingIndex": NewPortMappingIndex,
        }

        out_params = self.call_action("GetGenericPortMappingEntry", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewRemoteHost", "NewExternalPort", "NewProtocol", "NewInternalPort", "NewInternalClient", "NewEnabled", "NewPortMappingDescription", "NewLeaseDuration",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetIdleDisconnectTime(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetIdleDisconnectTime action.

            :returns: "NewIdleDisconnectTime"
        """
        arguments = { }

        out_params = self.call_action("GetIdleDisconnectTime", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewIdleDisconnectTime",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetListOfPortMappings(self, NewStartPort, NewEndPort, NewProtocol, NewManage, NewNumberOfPorts, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetListOfPortMappings action.

            :returns: "NewPortListing"
        """
        arguments = {
            "NewStartPort": NewStartPort,
            "NewEndPort": NewEndPort,
            "NewProtocol": NewProtocol,
            "NewManage": NewManage,
            "NewNumberOfPorts": NewNumberOfPorts,
        }

        out_params = self.call_action("GetListOfPortMappings", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewPortListing",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetNATRSIPStatus(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetNATRSIPStatus action.

            :returns: "NewRSIPAvailable", "NewNATEnabled"
        """
        arguments = { }

        out_params = self.call_action("GetNATRSIPStatus", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewRSIPAvailable", "NewNATEnabled",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetSpecificPortMappingEntry(self, NewRemoteHost, NewExternalPort, NewProtocol, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetSpecificPortMappingEntry action.

            :returns: "NewInternalPort", "NewInternalClient", "NewEnabled", "NewPortMappingDescription", "NewLeaseDuration"
        """
        arguments = {
            "NewRemoteHost": NewRemoteHost,
            "NewExternalPort": NewExternalPort,
            "NewProtocol": NewProtocol,
        }

        out_params = self.call_action("GetSpecificPortMappingEntry", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewInternalPort", "NewInternalClient", "NewEnabled", "NewPortMappingDescription", "NewLeaseDuration",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetStatusInfo(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetStatusInfo action.

            :returns: "NewConnectionStatus", "NewLastConnectionError", "NewUptime"
        """
        arguments = { }

        out_params = self.call_action("GetStatusInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewConnectionStatus", "NewLastConnectionError", "NewUptime",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetWarnDisconnectDelay(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetWarnDisconnectDelay action.

            :returns: "NewWarnDisconnectDelay"
        """
        arguments = { }

        out_params = self.call_action("GetWarnDisconnectDelay", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewWarnDisconnectDelay",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_RequestConnection(self, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RequestConnection action.
        """
        arguments = { }

        self.call_action("RequestConnection", arguments=arguments, aspects=aspects)

        return

    def action_RequestTermination(self, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RequestTermination action.
        """
        arguments = { }

        self.call_action("RequestTermination", arguments=arguments, aspects=aspects)

        return

    def action_SetAutoDisconnectTime(self, NewAutoDisconnectTime, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetAutoDisconnectTime action.
        """
        arguments = {
            "NewAutoDisconnectTime": NewAutoDisconnectTime,
        }

        self.call_action("SetAutoDisconnectTime", arguments=arguments, aspects=aspects)

        return

    def action_SetConnectionType(self, NewConnectionType, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetConnectionType action.
        """
        arguments = {
            "NewConnectionType": NewConnectionType,
        }

        self.call_action("SetConnectionType", arguments=arguments, aspects=aspects)

        return

    def action_SetIdleDisconnectTime(self, NewIdleDisconnectTime, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetIdleDisconnectTime action.
        """
        arguments = {
            "NewIdleDisconnectTime": NewIdleDisconnectTime,
        }

        self.call_action("SetIdleDisconnectTime", arguments=arguments, aspects=aspects)

        return

    def action_SetWarnDisconnectDelay(self, NewWarnDisconnectDelay, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetWarnDisconnectDelay action.
        """
        arguments = {
            "NewWarnDisconnectDelay": NewWarnDisconnectDelay,
        }

        self.call_action("SetWarnDisconnectDelay", arguments=arguments, aspects=aspects)

        return



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import socket
import threading
from akit.exceptions import AKitSemanticError

from akit.xthreading.looper import Looper
from akit.xthreading.looperpool import LooperPool
from akit.monitoring.reportmonitor import ReportMonitor
from akit.networking.interfaces import get_correspondance_ip_address


REPORT_HEADER_LENGTH = 7

class ReportingServiceLooper(Looper):
    """
    """

    def loop(self, packet) -> bool: # pylint: disable=no-self-use
        """
            Method that is overloaded by derived classes in order to implement a work loop.
        """
        service, ipaddr, rep_class, rep_topic, rep_content = packet

        monitor = service.lookup_monitor(rep_class, rep_topic)
        if monitor is not None:
            monitor.process_report(ipaddr, rep_content)

        return True


class ReportingService:
    """
    """

    def __init__(self, highwater: int=5, daemon=True):
        self._looper_pool = LooperPool(ReportingServiceLooper, min_loopers=1,
                                       max_loopers=1, highwater=highwater, daemon=daemon)

        self._daemon = daemon
        self._service_thread = None

        self._monitor_table = {}
        self._running = False

        self._port = 0
        self._svc_sock = None

        self._lock = threading.RLock()
        return

    def lookup_monitor(self, monitor_class: str, monitor_topic: str) -> ReportMonitor:

        monitor_name = "{}:{}".format(monitor_class, monitor_topic)

        monitor = None

        self._lock.acquire()
        try:
            if monitor_name in self._monitor_table:
                monitor = self._monitor_table[monitor_name]
        finally:
            self._lock.release()

        return monitor

    def register_monitor(self, monitor: ReportMonitor):
        """
        """
        monitor_name = "{}:{}".format(monitor.report_class, monitor.report_topic)

        self._lock.acquire()
        try:
            if monitor_name in self._monitor_table:
                errmsg = "A 'ReportMonitor' named '{}' has already been registered.".format(monitor_name)
                raise AKitSemanticError(errmsg) from None

            self._monitor_table[monitor_name] = monitor

            corr_ip = monitor.correspondance_ip
            monitor.set_report_endpoint(corr_ip, self._port)
        finally:
            self._lock.release()

        return

    def start(self):

        self._lock.acquire()
        try:
            sgate = threading.Event()
            sgate.clear()

            self._service_thread = threading.Thread(name="Accept", 
                target=self._service_thread_entry, args=(sgate,))
            self._service_thread.daemon = self._daemon
            self._service_thread.start()

            self._lock.release()
            try:
                sgate.wait()
            finally:
                self._lock.acquire()
        finally:
            self._lock.release()

        self._looper_pool.start_pool()

        return

    def _service_thread_entry(self, sgate):

        self._lock.acquire()
        try:
            self._running = True

            try:
                self._svc_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

                # Bind to all interfaces and also bint to port zero so we can get
                # an ephimeral port address
                self._svc_sock.bind(("0.0.0.0", 0))

                # We dont keep track of host because we need to provide a host address
                # that is relative to the device and the interface we can communicate
                # with a device on.
                _, self._port = self._svc_sock.getsockname()

                self._svc_sock.listen(1)

                sgate.set()
                sgate = None

                self._lock.release()
                try:

                    while self._running:
                        asock, claddr = self._svc_sock.accept()

                        try:
                            header = asock.recv(REPORT_HEADER_LENGTH)
                            msg_size = int(header.rstrip(b":"))

                            msg_body = asock.recv(msg_size)
                            msg_body = msg_body.decode()

                            rep_cls_end = msg_body.find(":")
                            rep_topic_end = msg_body.find(":", rep_cls_end + 1)
                            if rep_cls_end > -1 and rep_topic_end > -1:
                                rep_class = msg_body[:rep_cls_end]
                                rep_topic = msg_body[rep_cls_end + 1: rep_topic_end]
                                rep_content = msg_body[rep_topic_end + 1:]

                                wkpacket = (self, claddr, rep_class, rep_topic, rep_content)
                                self._looper_pool.push_work(wkpacket)

                        except:
                            import traceback
                            errmsg = traceback.format_exc()
                            print(errmsg)
                        finally:
                            asock.close()

                finally:
                    self._lock.acquire()

            finally:
                if self._svc_sock is not None:
                    self._svc_sock.close()

                if sgate is not None:
                    sgate.set()
        finally:
            self._lock.release()

        return




"""
.. module:: eventedvariablestate
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the :class:`EventedVariableState` enumeration represents
               the update status states of an evented variable.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2022, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from enum import IntEnum

class EventedVariableState(IntEnum):
    """
        An enumeration that indicates the state of the event variable.
    """
    UnInitialized = 0
    Default = 1
    Valid = 2
    Stale = 3



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class AddressBook1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:AddressBook:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'AddressBook1'

    SERVICE_DEFAULT_VARIABLES = {}

    SERVICE_EVENT_VARIABLES = {
        "IncomingRequest": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_Accept(self, RequestID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Accept action.
        """
        arguments = {
            "RequestID": RequestID,
        }

        self.call_action("Accept", arguments=arguments, aspects=aspects)

        return

    def action_FetchcontactInfo(self, Targetcontacts, ShareInfo, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the FetchcontactInfo action.
        """
        arguments = {
            "Targetcontacts": Targetcontacts,
            "ShareInfo": ShareInfo,
        }

        self.call_action("FetchcontactInfo", arguments=arguments, aspects=aspects)

        return

    def action_ImportContacts(self, NetworkAddressBookID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ImportContacts action.
        """
        arguments = {
            "NetworkAddressBookID": NetworkAddressBookID,
        }

        self.call_action("ImportContacts", arguments=arguments, aspects=aspects)

        return

    def action_Reject(self, RequestID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Reject action.
        """
        arguments = {
            "RequestID": RequestID,
        }

        self.call_action("Reject", arguments=arguments, aspects=aspects)

        return

    def action_RetrieveIncomingRequests(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RetrieveIncomingRequests action.

            :returns: "ActiveIncomingRequests"
        """
        arguments = { }

        out_params = self.call_action("RetrieveIncomingRequests", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("ActiveIncomingRequests",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_ShareContacts(self, SharedContacts, SharedInfo, TargetContacts, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ShareContacts action.
        """
        arguments = {
            "SharedContacts": SharedContacts,
            "SharedInfo": SharedInfo,
            "TargetContacts": TargetContacts,
        }

        self.call_action("ShareContacts", arguments=arguments, aspects=aspects)

        return

    def action_SharePCC(self, TargetContacts, ShareInfo, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SharePCC action.
        """
        arguments = {
            "TargetContacts": TargetContacts,
            "ShareInfo": ShareInfo,
        }

        self.call_action("SharePCC", arguments=arguments, aspects=aspects)

        return



"""
.. module:: upnpcoordinator
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Contains the UpnpCoordinator which is used for managing connectivity upnp devices visible in the automation landscape.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Dict, List, Optional, TYPE_CHECKING, Union

import os
import socket
import threading
import traceback
import uuid
import weakref

from http import HTTPStatus
from io import BytesIO, SEEK_END

import netifaces
import yaml

from requests.exceptions import ConnectionError

from akit.environment.variables import AKIT_VARIABLES
from akit.exceptions import AKitConfigurationError, AKitRuntimeError, AKitSemanticError, AKitTimeoutError
from akit.friendlyidentifier import FriendlyIdentifier

from akit.networking.constants import AKitHttpHeaders, HTTP1_1_LINESEP, HTTP1_1_END_OF_HEADER
from akit.networking.interfaces import get_interface_for_ip
from akit.networking.multicast import create_multicast_socket
from akit.networking.resolution import get_arp_table, refresh_arp_table
from akit.networking.trafficcapturecontext import TrafficCaptureContext

from akit.interop import upnp as upnp_module
from akit.interop.landscaping.landscapedevice import LandscapeDevice

from akit.interop.upnp.devices.upnprootdevice import UpnpRootDevice
from akit.interop.upnp.devices.upnprootdevice import device_description_load
from akit.interop.upnp.devices.upnprootdevice import device_description_find_components

from akit.interop.upnp.upnpfactory import UpnpFactory
from akit.interop.upnp.upnpprotocol import msearch_query_host, msearch_scan, notify_parse_request
from akit.interop.upnp.upnpprotocol import MSearchKeys, MSearchRouteKeys, UpnpProtocol
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy
from akit.interop.upnp.xml.upnpdevice1 import UPNP_DEVICE1_NAMESPACE

from akit.interop.coordinators.coordinatorbase import CoordinatorBase

from akit.networking.interfaces import get_ipv4_address

from akit.paths import get_path_for_output

# Types imported only for type checking purposes
if TYPE_CHECKING:
    from akit.interop.landscaping.landscape import Landscape

EMPTY_LIST = []

UPNP_DIR = os.path.dirname(upnp_module.__file__)

UPNP_SUBSCRIPTION_NOTIFY_RESPONSE_OK = HTTP1_1_LINESEP.join([
    b'HTTP/1.1 200 OK',
    b'Connection: close',
    b'Content-Length: 0',
    b'Content-Type: text/html',
    b'Server: %s,UPnP/1.0' % AKitHttpHeaders.SERVER.encode(),
    b''
])

def http_parse_header(header_content):
    headers = {}
    header_lines = header_content.splitlines(False)

    assert len(header_lines) > 0, "The header content was not a valid HTTP request header."
    req_line = header_lines[0]
    for hline in header_lines[1:]:
        name_end = hline.find(b":")
        if name_end > -1:
            hname = hline[:name_end].upper().decode().strip()
            hval = hline[name_end + 1:].decode().strip()
            headers[hname] = hval
    return req_line, headers


class UpnpCoordinator(CoordinatorBase):
    """
        The UpnpCoordinator utilizes the expected device declarations of type 'network/upnp' to establish and maintain connectivity
        and interoperability with UPNP based network devices.  The UpnpCoordinator will scan all interfaces except for the excluded
        interfaces.  It also creates a general broadcast monitor thread and a subscription dispatch thread for each interface it
        is monitoring.
    """
    # pylint: disable=attribute-defined-outside-init

    UPNP_CACHE_DIR = os.path.join(AKIT_VARIABLES.AKIT_HOME_DIRECTORY, "cache", "upnp")

    def __init__(self, lscape: "Landscape", control_point=None, workers: int = 5):
        super(UpnpCoordinator, self).__init__(lscape, control_point=control_point, workers=workers)

        self._local_udn = uuid.getnode()

        return

    def _initialize(self, *_args, control_point=None, workers: int = 5, **_kwargs):
        """
            Called by the CoordinatorBase constructor to perform the one time initialization of the coordinator Singleton
            of a given type.

            :param control_point: An object that acts as a UPNP control point.
            :param workers: The number of worker threads to spin up for handling work packets.
        """
        # pylint: disable=arguments-differ

        # ============================ Fixed Variables ============================
        # These variables are fixed at the start of the UpnpCoordinator and because
        # they are fixed, we don't protect them with the all.

        self._control_point = control_point
        self._worker_count = workers

        self._factory = UpnpFactory()

        self._running = False
        self._allow_unknown_devices = False
        self._upnp_recording = True

        # The count for the shutdown gate semaphore is set at startup so we don't
        # need to lock protect it.  Once it is set, it is fixed for the lifespan
        # of the UpnpCoordinator
        self._shutdown_gate = None

        # ======================= Coordinator Lock Variables =======================
        # These variables are protected and read/write synchronized by the coordinator
        # lock.  They are all prefixed with _cl_ so it is easy to identify in code
        # if the lock is being held properly when the variables are being accessed.

        # Callback threads are created on a per interface basis so we use a list
        # to manage them.
        self._cl_callback_threads = []

        # We don't want our threads that are answering requests or filtering traffic
        # to do extensive amounts of work, so we have a pool of worker threads that
        # that work is dispatched to for incoming requests and information processing.
        self._cl_worker_threads = []

        # A lookup table from USN -> devinfo for devices that were declared as
        # watched devices, the watched devices are devices that will be monitored
        # under special constraints and the coordinator will ensure they are found
        # during startup and will expend thread resources to ensure they are updated
        self._cl_watched_devices = {}

        self._cl_iface_callback_addr_lookup = None

        # A lookup table that store device registrations for differen subscription id(s)
        # The UpnpCoordinator spins up one thread per interface that needs to be monitored
        # for callback traffic from devices and provides a callback URL for subscriptions
        # to the devices,  the callback threads use this table to dispatch subscription
        # callbacks on given interfaces to the appropriate device. So UpnpEventVar instances
        # can be updated.
        self._cl_subscription_id_to_device = {}

        # A reverse lookup table that can lookup the device hint from its USN
        self._cl_usn_to_hint = {}

        # Callback capture management dictionaries
        self._cl_callback_traffic_capture_from = {}

        # =========================== Queue Lock Variables ==========================
        # Variables that manage the work queue and dispatching of work to worker threads
        self._queue_lock = threading.RLock()
        self._queue_available = threading.Semaphore(0)
        self._queue_work = []

        self._match_table = {
            "modelName": ('upnp', UpnpRootDevice._matches_model_name), # pylint: disable=protected-access
            "modelNumber": ('upnp', UpnpRootDevice._matches_model_number) # pylint: disable=protected-access
        }

        self._excluded_interfaces = ["lo"]

        self._cl_callback_interface_sockets = {}

        return

    @property
    def local_udn(self) -> str:
        """
            Returns a local UDN for this device.
        """
        return self._local_udn

    @property
    def watch_devices(self) -> List[LandscapeDevice]:
        """
            Returns a list of the devices being watched by the UPNP coordinator.
        """
        wlist = []

        self._coord_lock.acquire()
        try:
            wlist = [wd for wd in self._cl_watched_devices.values()]
        finally:
            self._coord_lock.release()

        return wlist

    def create_callback_traffic_capture_context(self, fromip: str):
        
        identifier = uuid.uuid4()
        context = TrafficCaptureContext(self, identifier, fromip, self.destroy_callback_traffic_capture_context)

        self._coord_lock.acquire()
        try:
            if fromip in self._cl_callback_traffic_capture_from:
                self._cl_callback_traffic_capture_from[fromip][context.identifier] = context
            else:
                self._cl_callback_traffic_capture_from[fromip] = {context.identifier: context}
        finally:
            self._coord_lock.release()

        return context

    def destroy_callback_traffic_capture_context(self, context: TrafficCaptureContext):

        fromip = context.fromip
        identifier = context.identifier

        self._coord_lock.acquire()
        try:
            if fromip in self._cl_callback_traffic_capture_from:
                fromip_table = self._cl_callback_traffic_capture_from[fromip]
                if identifier in fromip_table:
                    del fromip_table[identifier]
        finally:
            self._coord_lock.release()

        return

    def establish_presence(self):
        watched_devices = self.watch_devices

        for wdev in watched_devices:
            wdev.upnp.set_auto_subscribe(True, factory=self._factory)

        return

    def lookup_callback_url_for_interface(self, ifname: str) -> str:
        """
            The callback url that is utilized by devices to determine the URL that should be used as a callback
            for a given network interface.

            :param ifname: The interface name of the interface to get the callback url for.

            :returns: Returns the callback url being monintored by the coordinator for a given interface.
        """
        callback_url = None

        self._coord_lock.acquire()
        try:
            if ifname in self._cl_iface_callback_addr_lookup:
                callback_url = "<http://{}/notify>".format(self._cl_iface_callback_addr_lookup[ifname])
        finally:
            self._coord_lock.release()

        return callback_url

    def lookup_device_by_mac(self, mac: str) -> Union[LandscapeDevice, None]:
        """
            Lookup a UPNP device by its MAC address.

            :param mac: The mac address of the device to search for.

            :returns: The device found with the associated MAC address or None
        """

        found = None

        self._coord_lock.acquire()
        try:
            for nxtdev in self._cl_children.values():
                if mac == nxtdev.MACAddress:
                    found = nxtdev.basedevice
                    break
        finally:
            self._coord_lock.release()

        return found

    def lookup_device_by_upnp_hint(self, hint: str) -> Union[UpnpRootDevice, None]:
        """
            Lookup a UPNP device by its hint id.

            :param hint: The hint of the device to search for.

            :returns: The device found with the associated hint address or None
        """

        found = None

        self._coord_lock.acquire()
        try:
            for nxtdev in self._cl_children.values():
                if nxtdev.USN.find(hint) > -1:
                    found = nxtdev
                    break
            
            if found is None:
                errmsg_lines = [
                    "lookup_device_by_upnp_hint did not find a device for hint={}.".format(hint),
                    "DEVICE USN(s):"
                ]
                for nxtdev in self._cl_children.values():
                    errmsg_lines.append(nxtdev.USN)
                errmsg = os.linesep.join(errmsg_lines)
                self._logger.debug(errmsg)

        finally:
            self._coord_lock.release()

        return found
    
    def lookup_device_by_upnp_usn(self, usn: str) -> Union[UpnpRootDevice, None]:
        """
            Lookup a UPNP device by its USN id.

            :param usn: The USN of the device to search for.

            :returns: The device found with the associated USN address or None
        """

        found = None

        self._coord_lock.acquire()
        try:
            for nxtdev in self._cl_children.values():
                if nxtdev.USN.find(usn) > -1:
                    found = nxtdev
                    break
            
            if found is None:
                errmsg_lines = [
                    "lookup_device_by_upnp_usn did not find a device for usn={}.".format(usn),
                    "DEVICE USN(s):"
                ]
                for nxtdev in self._cl_children.values():
                    errmsg_lines.append(nxtdev.USN)
                errmsg = os.linesep.join(errmsg_lines)
                self._logger.debug(errmsg)

        finally:
            self._coord_lock.release()

        return found

    def lookup_device_list_by_usn(self, usnlist: List[str]) -> List[LandscapeDevice]:
        """
            Lookup a list of UPNP devices by thier USN id.

            :param usn: A list of USN identifiers for devices to search for.

            :returns: A list of devices found with USN(s) in the usn search list.
        """
        found = []

        self._coord_lock.acquire()
        try:
            for usn in usnlist:
                for nxtdev in self._cl_children.values():
                    if usn == nxtdev.USN:
                        found.append(nxtdev)
        finally:
            self._coord_lock.release()

        return found

    def lookup_device_instance_for_sid(self, sid: str) -> UpnpServiceProxy:
        """
            Lookup a device instance that has been registered with a subscription id.

            :param sid: The sid of a service device to search for.

            :returns: A device found to be associated with the sid provided.
        """
        device_inst = None

        self._coord_lock.acquire()
        try:
            if sid in self._cl_subscription_id_to_device:
                device_inst = self._cl_subscription_id_to_device[sid]
        finally:
            self._coord_lock.release()

        return device_inst

    def register_subscription_for_device(self, sid: str, device: LandscapeDevice):
        """
            Registers a service instance for event callbacks via a 'sid'.

            :param sid: The service subscription id to register a device with.
            :param device: The device to register as a handler for notifications for the sid provided.
        """

        self._coord_lock.acquire()
        try:
            self._cl_subscription_id_to_device[sid] = device
        finally:
            self._coord_lock.release()

        return

    def startup_scan(self, query_devices: Dict[str, dict], required_devices: Optional[List[str]] = None,
                           watchlist: Optional[List[str]] = None, exclude_interfaces: List = [],
                            response_timeout: float = 20, pre_msearch_timeout=20, retry: int = 2, 
                            upnp_recording: bool = False, allow_unknown_devices: bool = False):
        """
            Starts up and initilizes the UPNP coordinator by utilizing a hint list to determine
            what network interfaces to setup UPNP monitoring on.

            :param query_devices: A list of upnp USN(s) to search for.
            :param required_devices: A list of required USN(s) that we will setup watching and updates for.
            :param exclude_interfaces: Network interfaces that will be excluded from the search for devices.
            :param response_timeout: A timeout to wait for a response.
            :param retry: The number of retry attempts to make before giving up on finding devices.
            :param upnp_recording: Forces the updating or recording of device descriptions from devices found on the network.
        """
        # pylint: disable=dangerous-default-value

        self._allow_unknown_devices = allow_unknown_devices
        self._upnp_recording = upnp_recording
        self._excluded_interfaces = exclude_interfaces

        lscape = self.landscape

        if self._running:
            raise AKitRuntimeError("UpnpCoordinator.startup_scan called twice, The UpnpCoordinator is already running.") from None

        # Because we only allow this method to be called once, We don't need to lock the UpnpCoordinator
        # for most of this activity because the only thread with a reference to use is the caller.  At
        # the end of this function when we startup all the callback and worker threads is when we need to
        # start using the lock.
        if query_devices is None:
            query_devices = []

        interface_list = [ifname for ifname in netifaces.interfaces()]
        for exif in exclude_interfaces:
            if exif in interface_list:
                interface_list.remove(exif)

        found_devices = {}
        matching_devices = {}
        missing_devices = []

        remaining_query_devices = [ usn for usn in query_devices]

        pre_msearch_found = self._device_scan_pre_msearch_query(lscape, remaining_query_devices, pre_msearch_timeout, retry)
        if len(pre_msearch_found) > 0:
            found_devices.update(pre_msearch_found)
            matching_devices.update(pre_msearch_found)
            
            devices_to_remove = []
            for qdev_usn in remaining_query_devices:
                if qdev_usn in found_devices:
                    devices_to_remove.append(qdev_usn)
            
            # Remove the devices that we found and successfully queried from the list of device to query
            for fdev_usn in devices_to_remove:
                remaining_query_devices.remove(fdev_usn)

        # ======================== M-SEARCH ========================
        # We have to first attempt to msearch for the devices because we need the latest
        # device info to create devices with.
        if len(remaining_query_devices) > 0:
            msearch_found, msearch_matching = self._device_msearch_scan(
                remaining_query_devices, interface_list=interface_list,
                response_timeout=response_timeout, retry=retry)

            found_devices.update(msearch_found)
            matching_devices.update(msearch_matching)

            # Remove the devices that we found from the list of device to query for
            for match_hint in msearch_matching:
                if match_hint in remaining_query_devices:
                    remaining_query_devices.remove(match_hint)

        # ====================== UPNP CACHE & DIRECT QUERY ===================
        # If we didn't find a device with an M-SEARCH, look in the UPNP cache
        # and try to query the device directly to get its info.  We only use the
        # cached info to get the IP of the device, we don't use it for other data
        # because some of the data might be out of date.
        if len(remaining_query_devices) > 0:
            # When we look in the cache, the found devices are also the matching devices
            cache_found = self._device_cache_scan(remaining_query_devices)
            found_devices.update(cache_found)
            matching_devices.update(cache_found)

            devices_to_remove = []
            for qdev_usn in remaining_query_devices:
                if qdev_usn in cache_found:
                    cdev_info = cache_found[qdev_usn]
                    cached_ipaddr = cdev_info[MSearchKeys.IP]
                    qdev_info = msearch_query_host(qdev_usn, cached_ipaddr)
                    if qdev_info is not None:
                        found_devices[fdev_usn] = qdev_info
                        devices_to_remove.append(fdev_usn)

            # Remove the devices that we found and successfully queried from the list of device to query
            for fdev_usn in devices_to_remove:
                remaining_query_devices.remove(fdev_usn)

        # ====================== LAST RESULT, TRY USING THE ARP TABLE ===================
        if len(remaining_query_devices) > 0:
            final_found, final_matching = self._device_scan_final_alternatives(remaining_query_devices,
                interface_list=interface_list, response_timeout=response_timeout,
                retry=retry)

        missing_devices = []
        if required_devices is not None:
            for devhint in required_devices:
                if devhint not in matching_devices:
                    missing_devices.append(devhint)

        if len(missing_devices) > 0:
            errmsg_list = [
                "Error devices missing from configuration.",
                "MISSING DEVICES:"
            ]
            for devhint in missing_devices:
                errmsg_list.append("    %s" % devhint)
            errmsg = os.linesep.join(errmsg_list)
            self._logger.error(errmsg)
            raise AKitConfigurationError(errmsg) from None

        config_lookup = lscape._internal_get_upnp_device_config_lookup_table() # pylint: disable=protected-access

        for dhint, dval in matching_devices.items():
            addr = dval[MSearchKeys.IP]
            location = dval[MSearchKeys.LOCATION]
            self._update_root_device(lscape, config_lookup, addr, location, dhint, dval)

        if watchlist is not None and len(watchlist) > 0:
            for hint in watchlist:
                wdev = self.lookup_device_by_upnp_hint(hint)
                self._cl_watched_devices[hint] = wdev.basedevice

        self._log_scan_results(found_devices, matching_devices, missing_devices)

        self._device_cache_update(matching_devices)

        self._start_all_threads()

        self._found_devices = found_devices
        self._available_devices = matching_devices
        self._unavailable_devices = missing_devices

        return found_devices, matching_devices, missing_devices

    def wakeup_device(self, usn, dev_hint):
        self._logger.info("TODO: Implement an overload for UpnpCoordinatorIntegration:wakeup_device.")
        return

    def _create_root_device(self, manufacturer: str, model_number: str, model_description: str) -> UpnpRootDevice:
        """
            Create a root UPNP device using the provided manufacturer and modelNumber provided.  If no
            device is found for the provided manufacturer and model_number then a generic UPNP root device
            will be instantiated using the manufacturer, model_number and model_description.

            :param manufacturer: The manufacturer of the root device to instantiate.
            :param model_number: The model_number of the root device to instantiate.
            :param model_description: The description to give to a device when it is instantiated.

            :returns: A upnp root device instance.
        """
        dev = self._factory.create_root_device_instance(manufacturer, model_number, model_description)
        return dev

    def _device_cache_scan(self, query_devices: list):

        found_devices = {}

        for qdev_hint in query_devices:
            qdev_filename = os.path.join(self.UPNP_CACHE_DIR, qdev_hint)

            if os.path.exists(qdev_filename):
                dev_info = None
                with open(qdev_filename, 'r') as qdf:
                    dicontent = qdf.read()
                    dev_info = yaml.safe_load(dicontent)

                if dev_info is not None:
                    verified = self._device_cache_verify_device_info(qdev_hint, qdev_filename, dev_info)
                    if verified:
                        found_devices[qdev_hint] = dev_info

        return found_devices

    def _device_cache_update(self, found_device: dict):

        if not os.path.exists(self.UPNP_CACHE_DIR):
            os.makedirs(self.UPNP_CACHE_DIR)

        for fdev_usn in found_device:
            fdev_info = found_device[fdev_usn]
            fdev_filename = os.path.join(self.UPNP_CACHE_DIR, fdev_usn)
            with open(fdev_filename, 'w') as dcf:
                yaml.dump(fdev_info, dcf)

        return

    def _device_cache_verify_device_info(self, device_hint: str, device_filename: str, device_info: dict):
        valid = False

        if MSearchKeys.USN_DEV in device_info and "LOCATION" in device_info:
            location_url = device_info["LOCATION"]

            try:
                dev_desc = device_description_load(location_url)
                if dev_desc is not None:
                    namespaces = {"": UPNP_DEVICE1_NAMESPACE}

                    root_element = dev_desc.getroot()
                    udn_element = root_element.find("device/UDN", namespaces=namespaces)
                    if udn_element is not None:
                        udn_full = udn_element.text.lstrip("uuid:")
                        if udn_full.find(device_hint) < 0:
                            os.remove(device_filename)
                        else:
                            valid = True
            except ConnectionError:
                pass
        else:
            os.remove(device_filename)

        return valid

    def _device_msearch_scan(self, query_devices: list, interface_list: list, response_timeout: float = 20, 
                             retry: int = 2, custom_headers: Optional[dict]=None):

        found_devices = {}
        matching_devices = {}

        show_progress = False
        if self._interactive_mode:
            show_progress = True

        remaining_to_find = [hint for hint in query_devices]

        for ridx in range(0, retry):
            if ridx > 0:
                self._logger.info("MSEARCH: Not all devices found, retrying (count=%d)..." % ridx)
            iter_found_devices, iter_matching_devices = msearch_scan(remaining_to_find,
                interface_list=interface_list, response_timeout=response_timeout, custom_headers=custom_headers,
                show_progress=show_progress)
            found_devices.update(iter_found_devices)
            matching_devices.update(iter_matching_devices)

            for devusn in matching_devices.keys():
                if devusn in remaining_to_find:
                    remaining_to_find.remove(devusn)

            if len(remaining_to_find) == 0:
                break

        return found_devices, matching_devices

    def _device_scan_final_alternatives(self, query_devices: list, interface_list: list, response_timeout: float = 20, retry: int = 2):

        found_devices = {}
        matching_devices = {}

        missing_devices_checklist = [d for d in query_devices]

        # If we are missing some devices, lets try to wake them up and also attempt to find
        # them in the ARP table.
        self._logger.info("* Forcing the ARP table to refresh.")
        refresh_arp_table()

        # Try to use the ARP table to hint about the device
        self._logger.info("* Looking for device and interface hints in the APR table.")
        requery_devices = []
        arp_table = get_arp_table(normalize_hwaddr=True)
        for dmac, dinfo in arp_table.items():
            usn_match = None
            for usn in missing_devices_checklist:
                usn_norm = usn.replace(":", "").upper()
                if usn_norm.find(dmac) > -1:
                    usn_match = usn
                    break

            if usn_match is not None:
                missing_devices_checklist.remove(usn_match)
                requery_devices.append((usn_match, dinfo))

        self._logger.info("* Trying to requery the missing devices.")
        # As a last resort, rescan for the missing devices directly via unicast.
        for expusn, dev_hint in requery_devices:
            try:
                self.wakeup_device(expusn, dev_hint)

                requery_ip = None
                if dev_hint is not None:
                    requery_ip=dev_hint["ip"]

                device_info = msearch_query_host(requery_ip, query_usn=expusn, response_timeout=response_timeout)
                if device_info is not None:
                    found_devices[expusn] = device_info
                    query_devices.remove(expusn)

            except AKitTimeoutError:
                pass

        return found_devices, matching_devices

    def _device_scan_pre_msearch_query(self, lscape: "Landscape", query_devices: list, timeout: float, interval: float):
        """
            This method can be used by derived coordinators to find devices via other protocols
            prior to the use of the msearch protocol.
        """
        return []

    def _log_scan_results(self, found_devices: dict, matching_devices:dict , missing_devices: list):
        """
            Logs the results of the device scan.

            :param found_devices: A list of found devices.
            :param matching_devices: A list of devices matching the expected devices listed in the upnp_hint_list.
            :param missing_devices: A list of USN(s) for devices that were not found.
        """
        devmsg_lines = ["FOUND DEVICES:"]
        for dkey, _ in found_devices.items():
            devmsg_lines.append("    %s" % dkey)
        devmsg_lines.append("")

        devmsg_lines.append("MATCHING DEVICES:")
        for dkey, _ in matching_devices.items():
            devmsg_lines.append("    %s" % dkey)
        devmsg_lines.append("")

        if len(missing_devices) > 0:
            devmsg_lines.append("MISSING DEVICES:")
            for dkey in missing_devices:
                devmsg_lines.append("    %s" % dkey)
            devmsg_lines.append("")

        devmsg = os.linesep.join(devmsg_lines)
        self._logger.info(devmsg)

        return

    def _normalize_name(self, name: str) -> str:
        """
            Normalizes a name by removing non-alphanumeric characters.

            :param name: A name to normalize.

            :returns: The normalized name with all the non-alphanumeric characters removed.
        """
        # pylint: disable=no-self-use

        normal_chars = [ nc for nc in name if str.isalnum(nc) ]
        normal_name = ''.join(normal_chars)

        return normal_name

    def _process_device_alive_notification(self, ip_addr, usn_device: str, host: str, request_info: dict):
        """
            Processes a device alive notification when given the USN and dictionary of header keys and values.

            :param usn: The USN of the device which the notification is from.
            :param request_info: The headers from the notification request.
        """

        ip_addr = request_info[MSearchKeys.IP]
        location = request_info[MSearchKeys.LOCATION]

        # Get a reference to the landscape
        lscape = self.landscape

        self._activate_root_device(lscape, usn_device, ip_addr, location, request_info)

        return

    def _process_device_byebye_notification(self, ip_addr, usn_device: str, host: str, request_info: dict):
        """
            Processes a device byebye notification when given the USN and dictionary of header keys and values.

            :param usn: The USN of the device which the notification is from.
            :param request_info: The headers from the notification request.
        """

        ip_addr = request_info[MSearchKeys.IP]

        location = "Unknown"
        if MSearchKeys.LOCATION in request_info:
            location = request_info[MSearchKeys.LOCATION]

        # Get a reference to the landscape
        lscape = self.landscape

        self._deactivate_root_device(lscape, usn_device, ip_addr, location, request_info)

        return

    def _process_service_alive_notification(self, ip_addr, usn_device: str, usn_class: str, host: str, request_info: dict):
        return

    def _process_service_byebye_notification(self, ip_addr, usn_device: str, usn_class: str, host: str, request_info: dict):
        return

    def _process_subscription_callback(self, ifname: str, claddr: str, asock: socket):
        """
            Processes a callback request on the specified interface and address request.

            :param ifname: The name of the ifname the request came from.
            :param claddr: The address of the client that made the callback.
            :param request: The content of the callback request.
        """
        # pylint: disable=unused-argument

        cbbuffer = BytesIO()

        req_headers = None

        headers_length = None
        content_length = None

        message_length = None
        read_so_far = 0
        remaining_length = 1024
        try:
            # Read To Header
            while self._running and remaining_length > 0:
                # Process the requests
                nxtbuff = asock.recv(remaining_length)
                read_so_far += len(nxtbuff)

                # If we didn't get anything from recv, it means we hit the end of
                # the pipe.  We can exit the read loop.
                if len(nxtbuff) == 0:
                    break

                cbbuffer.write(nxtbuff)

                if req_headers is None:
                    current_content = cbbuffer.getvalue()
                    header_end = current_content.find(HTTP1_1_END_OF_HEADER)
                    if header_end > -1:
                        header_content = current_content[:header_end]
                        headers_length = len(header_content) + len(HTTP1_1_END_OF_HEADER)
                        req_line, req_headers = http_parse_header(header_content)
                        if "CONTENT-LENGTH" in req_headers:
                            content_length = int(req_headers["CONTENT-LENGTH"])
                            message_length = headers_length + content_length
                
                if message_length is not None:
                    remaining_length = message_length - read_so_far

        finally:
            asock.sendall(UPNP_SUBSCRIPTION_NOTIFY_RESPONSE_OK)
            asock.close()

        request = cbbuffer.getvalue()

        cbbuffer.truncate(0)
        cbbuffer.seek(0)

        req_body = request[headers_length:]

        capture_to = None
        self._coord_lock.acquire()
        try:
            if claddr in self._cl_callback_traffic_capture_from:
                capture_to = [cc for cc in self._cl_callback_traffic_capture_from[claddr].values()]
        finally:
            self._coord_lock.release()

        if capture_to is not None:
            for cc in capture_to:
                cc.append_capture(req_headers, req_body)

        if "SID" in req_headers:
            sid = req_headers["SID"]

            device = None
            #lookup the device that needs to handle this subscription
            self._coord_lock.acquire()
            try:
                if sid in self._cl_subscription_id_to_device:
                    device = self._cl_subscription_id_to_device[sid]
            finally:
                self._coord_lock.release()

            if device is not None:
                device.process_subscription_callback(claddr, sid, req_headers, req_body)
            else:
                errmsg = "Received subscription notification for unknown sid.  claddr={} sid={}".format(claddr, sid)
                self._logger.error(errmsg)
        else:
            dbgmsg_lines = [
                "Received subscription notification without an SID. claddr={}".format(claddr),
                "REQUEST HEADERS:"
            ]

            for hdr_key, hdr_val in req_headers.items():
                dbgmsg_lines.append("    {}:{}".format(hdr_key, hdr_val))

            dbgmsg_lines.append("REQUEST BODY:")
            dbgmsg_lines.append(req_body.decode("utf-8"))

            dbgmsg = os.linesep.join(dbgmsg_lines)

            self._logger.debug(dbgmsg)

        return

    def _process_request_for_msearch(self, addr: str, request: str):
        """
            Process an msearch request from the given address.

            :param addr: The address of machine making the msearch request.
            :param request: The request content of the msearch request to process.
        """
        #reqinfo = msearch_parse_request(request)
        self._logger.debug("RESPONDING TO MSEARCH")

        return

    def _process_request_for_notify(self, addr: str, request: str):
        """
            Process a notify request.

            :param addr: The address of machine making the notify request.
            :param request: The request content of the notify request to process.
        """
        req_headers, _ = notify_parse_request(request)

        usn = req_headers["USN"]
        host = req_headers["HOST"]
        subtype = req_headers["NTS"]

        ip_addr, _ = addr
        if MSearchKeys.IP not in req_headers:
            req_headers[MSearchKeys.IP] = ip_addr

        if usn.find("::") > -1:
            usn_device, usn_class = usn.split("::")
            usn_device = usn_device.lstrip("uuid:")

            ifname = get_interface_for_ip(ip_addr)

            req_headers[MSearchKeys.USN_DEV] = usn_device
            req_headers[MSearchKeys.USN_CLS] = usn_class
            req_headers[MSearchKeys.ROUTES] = {
                ifname:
                {
                    MSearchRouteKeys.IFNAME: ifname,
                    MSearchRouteKeys.IP: ip_addr
                }
            }

            if subtype == "ssdp:alive":
                if usn_class == "upnp:rootdevice":
                    self._process_device_alive_notification(ip_addr, usn_device, host, req_headers)
                else:
                    self._process_service_alive_notification(ip_addr, usn_device, usn_class, host, req_headers)
            elif subtype == "ssdp:byebye":
                if usn_class == "upnp:rootdevice":
                    self._process_device_byebye_notification(ip_addr, usn_device, host, req_headers)
                else:
                    self._process_service_byebye_notification(ip_addr, usn_device, usn_class, host, req_headers)
            else:
                self._logger.info("UNKNOWNK NOTIFY - USN: %s HOST: %s SUBTYPE: %s" % (usn, host, subtype))

        return

    def _start_all_threads(self):
        """
            Starts up all the thread the UPNP coordinator uses for monitoring, callback notification servicing and
            the worker threads used for processing work packges from notifications and callbacks.
        """

        ifacelist = []
        for wdev in self.watch_devices:
            primary_route = wdev.upnp.primary_route
            ifname = primary_route[MSearchRouteKeys.IFNAME]
            if ifname not in ifacelist:
                ifacelist.append(ifname)
        ifacecount = len(ifacelist)

        self._coord_lock.acquire()
        try:
            sgate = threading.Event()

            self._running = True

            self._shutdown_gate = threading.Semaphore(self._worker_count + ifacecount + 1)

            # Spin-up the worker threads first so they will be ready to handle work packets
            for wkrid in range(0, self._worker_count):
                sgate.clear()
                wthread = threading.Thread(name="UpnpCoordinator - Worker(%d)" % wkrid, target=self._thread_entry_worker,
                                                   daemon=True, args=(sgate,))
                self._cl_worker_threads.append(wthread)

                # Release the coordinator lock while starting up the thread so we are not holding it
                # so the new thread can get access to register itself with internal UpnpCoordinator state
                self._coord_lock.release()
                try:
                    wthread.start()
                    sgate.wait()
                finally:
                    self._coord_lock.acquire()

            # Spin-up the Monitor thread so it can monitor notification traffic
            sgate.clear()
            self._monitor_thread = threading.Thread(name="UpnpCoordinator - Monitor", target=self._thread_entry_monitor,
                                                   daemon=True, args=(sgate,))
            self._monitor_thread.start()
            sgate.wait()

            # Spin-up a Callback thread for each interface
            self._cl_iface_callback_addr_lookup = {}

            for ifaceidx in range(0, ifacecount):
                ifname = ifacelist[ifaceidx]
                sgate.clear()
                cbthread = threading.Thread(name="UpnpCoordinator - Callback(%s)" % ifname, target=self._thread_entry_callback,
                                                    daemon=True, args=(sgate, ifname))
                self._cl_callback_threads.append(cbthread)

                # Release the coordinator lock while starting up the thread so we are not holding it
                # so the new thread can get access to register itself with internal UpnpCoordinator state
                self._coord_lock.release()
                try:
                    cbthread.start()
                    sgate.wait()
                finally:
                    self._coord_lock.acquire()

        except:
            self._running = False

            for _ in range(0, self._worker_count + ifacecount + 1):
                self._shutdown_gate.release()

            raise
        finally:
            self._coord_lock.release()

        return

    def _thread_entry_callback(self, sgate: threading.Event, ifname: str):
        """
            The entry point for callback processing threads.

            :param sgate: The startup gate that is used to indicate to the thread spinning up the coordinator
                          that the callback thread has started.
            :param ifname: The interface name of interface this thread is assigned to process callbacks for.
        """
        self._shutdown_gate.acquire()

        try:
            service_addr = ""

            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

            # Bind to port zero so we can get an ephimeral port address
            sock.bind((service_addr, 0))

            _, port = sock.getsockname()
            host = get_ipv4_address(ifname)

            iface_callback_addr = "%s:%s" % (host, port)

            self._cl_iface_callback_addr_lookup[ifname] = iface_callback_addr
            self._cl_callback_interface_sockets[ifname] = sock

            # Set the start gate to allow the thread spinning us up to continue
            sgate.set()

            sock.listen(1)

            while self._running:
                asock, claddr = sock.accept()

                # Queue the callback workpacket for dispatching by a worker thread
                wkpacket = (self._process_subscription_callback, (ifname, claddr, asock))

                self._queue_lock.acquire()
                try:
                    self._queue_work.append(wkpacket)
                    self._queue_available.release()
                finally:
                    self._queue_lock.release()

        finally:
            self._shutdown_gate.release()

        return

    def _thread_entry_monitor(self, sgate: threading.Event):
        """
            The entry point for monitor thread.  The monitor thread sets up monitoring on all interfaces.

            :param sgate: The startup gate that is used to indicate to the thread spinning up the coordinator
                          that the monitor thread has started.
        """
        self._shutdown_gate.acquire()

        multicast_address = UpnpProtocol.MULTICAST_ADDRESS
        multicast_port = UpnpProtocol.PORT

        try:
            # Set the start gate to allow the thread spinning us up to continue
            sgate.set()

            while self._running:

                sock = create_multicast_socket(multicast_address, multicast_port)

                try:
                    request, addr = sock.recvfrom(1024)

                    if request.startswith(b"M-SEARCH"):
                        if self._control_point:
                            wkpacket = (self._process_request_for_msearch, (addr, request))
                            self._queue_lock.acquire()
                            try:
                                self._queue_work.append(wkpacket)
                                self._queue_available.release()
                            finally:
                                self._queue_lock.release()
                    elif request.startswith(b"NOTIFY"):
                        wkpacket = (self._process_request_for_notify, (addr, request))
                        self._queue_lock.acquire()
                        try:
                            self._queue_work.append(wkpacket)
                            self._queue_available.release()
                        finally:
                            self._queue_lock.release()
                    else:
                        dbgmsg = b"UNKNOWN REQUEST TYPE:\n" + request
                        dbgmsg = dbgmsg.decode("utf-8")
                        self._logger.debug(dbgmsg)

                finally:
                    sock.close()

        finally:
            self._shutdown_gate.release()

        return

    def _thread_entry_worker(self, sgate: threading.Event):
        """
            The entry point for worker threads.  Worker threads process work packets queued up by other threads.

            :param sgate: The startup gate that is used to indicate to the thread spinning up the coordinator
                          that this worker thread has started.
        """
        try:
            # Set the start gate to allow the thread spinning us up to continue
            sgate.set()

            while self._running:
                self._queue_available.acquire()

                self._queue_lock.acquire()
                try:
                    wkfunc, wkargs = self._queue_work.pop(0)
                    wkfunc(*wkargs)
                except: # pylint: disable=bare-except
                    self._logger.exception("UpnpCoordinator: Worker exception.")
                finally:
                    self._queue_lock.release()

        finally:
            self._shutdown_gate.release()

        return

    def _activate_root_device(self, lscape: "Landscape", usn_device: str, ip_addr: str, location: str, deviceinfo: dict):
        """
        """

        dev = self.lookup_device_by_upnp_usn(usn_device)
        if dev is not None:
            # Mark the device active
            dev.mark_alive()
        elif self._allow_unknown_devices:
            config_lookup = lscape._internal_get_upnp_device_config_lookup_table() # pylint: disable=protected-access

            usn_dev = deviceinfo[MSearchKeys.USN_DEV]

            marked_as_skip = False
            if usn_dev in config_lookup:
                configinfo = config_lookup[usn_dev]
                if "skip" in configinfo and configinfo["skip"]:
                    marked_as_skip = True

            if not marked_as_skip:
                devhint = None
                self._coord_lock.acquire()
                try:
                    devhint = self._cl_usn_to_hint[usn_device]
                finally:
                    self._coord_lock.release()

                self._update_root_device(lscape, config_lookup, ip_addr, location, devhint, deviceinfo)

        return

    def _deactivate_root_device(self, lscape: "Landscape", usn_device: str, ip_addr: str, location: str, deviceinfo: dict):
        """
        """
        dev = self.lookup_device_by_upnp_usn(usn_device)
        if dev is not None:
            # Mark the device active
            dev.mark_byebye()

        return

    def _update_root_device(self, lscape, config_lookup: dict, ip_addr: str, location: str, devhint: str, deviceinfo: dict):
        """
            Updates a UPNP root device.

            :param lscape: The landscape singleton instance.
            :param config_lookup: A configuration lookup table that can be used to lookup configuration information for upnp devices.
            :param ip_addr: The IP address of the device to update.
            :param location: The location URL associated with the device.
            :param deviceinfo: The device information from the msearch response headers.
            :param devhint: The identifier hint used to identify the device.
        """

        if MSearchKeys.USN_DEV in deviceinfo:
            try:
                usn_dev = deviceinfo[MSearchKeys.USN_DEV]
                usn_cls = deviceinfo[MSearchKeys.USN_CLS]

                configinfo = None
                if usn_dev in config_lookup:
                    configinfo = config_lookup[usn_dev]

                docTree = device_description_load(location)

                try:
                    # {urn:schemas-upnp-org:device-1-0}root
                    namespaces = {"": UPNP_DEVICE1_NAMESPACE}

                    deviceDescParts = device_description_find_components(location, docTree, namespaces=namespaces)
                    devNode, urlBase, manufacturer, modelName, modelNumber, modelDescription = deviceDescParts

                    dev_extension = None
                    skip_device = False
                    try:
                        # Acquire the lock before we decide if the location exists in the children table
                        self._coord_lock.acquire()
                        if location not in self._cl_children:
                            try:
                                # Unlock while we do some expensive stuff, we have already decided to
                                # create the device
                                self._coord_lock.release()

                                try:
                                    # We create the device
                                    dev_extension = self._create_root_device(manufacturer, modelNumber, modelDescription)
                                except:
                                    errmsg = "ERROR: Unable to create device mfg=%s model=%s desc=%s\nTRACEBACK:\n" % (manufacturer, modelNumber, modelDescription)
                                    errmsg += traceback.format_exc()
                                    self._logger.error(errmsg)
                                    raise

                                if isinstance(dev_extension, UpnpRootDevice):
                                    dev_extension.record_description(ip_addr, urlBase, manufacturer, modelName, docTree, devNode, namespaces=namespaces, upnp_recording=self._upnp_recording)

                                coord_ref = weakref.ref(self)

                                basedevice: LandscapeDevice = lscape._internal_lookup_device_by_hint(devhint)
                                if basedevice is None:
                                    if self._allow_unknown_devices:
                                        dev_type = "network/upnp"
                                        dev_config_info = {
                                            "deviceType": dev_type,
                                            "deviceClass": "unknown",
                                            "upnp": {
                                                "hint": usn_dev,
                                                "modelName": modelName,
                                                "modelNumber": modelNumber
                                            },
                                            "USN_DEV": usn_dev,
                                            "USN_CLS": usn_cls
                                        }
                                        dfid = FriendlyIdentifier(usn_dev, devhint)
                                        basedevice = lscape._create_landscape_device(dfid, dev_type, dev_config_info)
                                    else:
                                        skip_device = True

                                if basedevice is not None:
                                    basedevice.update_full_identifier(usn_dev)
                                    basedevice = lscape._enhance_landscape_device(basedevice, dev_extension)

                                    basedevice_ref = weakref.ref(basedevice)

                                    dev_extension.initialize(coord_ref, basedevice_ref, usn_dev, location, configinfo, deviceinfo)

                                    # Refresh the description
                                    dev_extension.refresh_description(ip_addr, self._factory, docTree.getroot(), namespaces=namespaces)
                                    dev_extension.mark_alive()

                                    basedevice.attach_extension("upnp", dev_extension)

                                    basedevice.initialize_features()
                                    basedevice.update_match_table(self._match_table)

                                    lscape._internal_activate_device(devhint)
                            finally:
                                self._coord_lock.acquire()

                            if skip_device:
                                infomsg = "Skipping device usn={} location={}".format(usn_dev, location)
                                self._logger.info(infomsg)
                            # If the device is still not in the table, add it
                            elif location not in self._cl_children:
                                self._cl_children[location] = dev_extension

                            self._cl_usn_to_hint[usn_dev] = devhint

                        else:
                            dev_extension = self._cl_children[location]
                            # Refresh the description
                            dev_extension.refresh_description(ip_addr, self._factory, docTree.getroot(), namespaces=namespaces)
                    finally:
                        self._coord_lock.release()

                except AKitSemanticError:
                    # Always allow semantic errors to propagate, semantic errors represent
                    # misuse of a test framework API and can be definitively classified
                    # as a test code bug that needs to be fixed.
                    raise

                except:  # pylint: disable=bare-except
                    exmsg = traceback.format_exc()
                    errmsg_lines = [
                        "ERROR: Unable to parse description for. IP: %s LOCATION: %s" % (ip_addr, location),
                        exmsg
                    ]
                    for k, v in deviceinfo.items():
                        errmsg_lines.append("    %s: %s" % (k, v))

                    errmsg = os.linesep.join(errmsg_lines)
                    self._logger.debug(errmsg)

            except AKitSemanticError:
                # Always allow semantic errors to propagate, semantic errors represent
                # misuse of a test framework API and can be definitively classified
                # as a test code bug that needs to be fixed.
                raise

            except:  # pylint: disable=bare-except
                exmsg = traceback.format_exc()
                errmsg_lines = [
                    "ERROR: Unable to parse description for. IP: %s LOCATION: %s" % (ip_addr, location),
                    exmsg
                ]
                for k, v in deviceinfo.items():
                    errmsg_lines.append("    %s: %s" % (k, v))

                errmsg = os.linesep.join(errmsg_lines)
                self._logger.debug(errmsg)

        return



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Any, Iterable

def compare_any_of_same_type(expected: Any, found: Any) -> bool:
    result = None

    exp_type = type(expected)
    fnd_type = type(found)

    if exp_type != fnd_type:
        errmsg = "The types of the expected and found parameters be identical. exptype={} fndtype={}".format(exp_type, fnd_type)
        raise ValueError(errmsg)

    if exp_type == dict:
        compare_dict(expected, found)
    elif exp_type == list or exp_type == tuple:
        compare_iterables(expected, found)
    else:
        result = expected == found

    return result

def compare_dict(expected: dict, found: dict) -> bool:
    
    result = True

    if len(found) != len(expected):
        result = False

    else:
        for expkey, expval in expected.items():
            if expkey not in found:
                result = False
                break

            fndval = found[expkey]
            if type(expval) != type(fndval):
                result = False
                break

            result = compare_any_of_same_type(expval, fndval)
            if not result:
                break

    return result

def compare_iterables(expected: Iterable, found: Iterable) -> bool:

    result = True

    if len(found) != len(expected):
        result = False

    else:
        for eitem in expected:
            if eitem not in found:
                result = False
                break

    return result



"""
.. module:: sshpoolcoordinator
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Contains the SshPoolCoordinator which is used for managing connectivity with pools of ssh capable devices

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import List, Optional, Union, TYPE_CHECKING

import os
import pprint
import socket
import weakref

from akit.exceptions import AKitConfigurationError
from akit.friendlyidentifier import FriendlyIdentifier

from akit.interop.coordinators.coordinatorbase import CoordinatorBase
from akit.interop.coordinators.upnpcoordinator import UpnpCoordinator

from akit.interop.landscaping.landscapedevice import LandscapeDevice

from akit.interop.agents.sshagent import SshAgent

if TYPE_CHECKING:
    from akit.interop.landscaping.landscape import Landscape

def format_ssh_device_configuration_error(message, sshdev_config):
    """
        Takes an error message and an SSH device configuration info dictionary and
        formats a configuration error message.
    """
    error_lines = [
        message,
        "DEVICE:"
    ]

    dev_repr_lines = pprint.pformat(sshdev_config, indent=4).splitlines(False)
    for dline in dev_repr_lines:
        error_lines.append("    " + dline)
    
    errmsg = os.linesep.join(error_lines)
    return errmsg

class SshPoolCoordinator(CoordinatorBase):
    """
        The :class:`SshPoolCoordinator` creates a pool of agents that can be used to
        coordinate the interop activities of the automation process and remote SSH
        nodes.
    """
    # pylint: disable=attribute-defined-outside-init

    def __init__(self, lscape: "Landscape", *args, **kwargs):
        super(SshPoolCoordinator, self).__init__(lscape, *args, **kwargs)
        return

    def _initialize(self, *_args, **_kwargs):
        """
            Called by the CoordinatorBase constructor to perform the one time initialization of the coordinator Singleton
            of a given type.
        """
        # pylint: disable=arguments-differ
        self._cl_upnp_hint_to_ip_lookup = {}
        self._cl_ip_to_host_lookup = {}
        return

    def attach_to_devices(self, sshdevices: List[LandscapeDevice], upnp_coord: Optional[UpnpCoordinator]=None):
        """
            Processes a list of device configs and creates and registers devices and SSH device extensions
            attached with the landscape for the devices not already registered.  If a device has already
            been registered by the UPNP coordinator then a device extension is created and attached to the
            existing device.

            :param sshdevices: A list of ssh device configuration dictionaries.
            :param upnp_coord: The UpnpCoordinator singleton instance.
        """

        lscape = self.landscape

        ssh_config_errors = []

        ssh_devices_available = []
        ssh_devices_unavailable = []

        for sshdev in sshdevices:
            sshdev_config = sshdev.device_config
            ssh_credential_list = sshdev.ssh_credentials
            if len(ssh_credential_list) == 0:
                errmsg = format_ssh_device_configuration_error(
                        "All SSH devices must have at least one valid credential.", sshdev_config)
                raise AKitConfigurationError(errmsg) from None

            ssh_cred_by_role = {}
            for ssh_cred in ssh_credential_list:
                cred_role = ssh_cred.role
                if cred_role not in ssh_cred_by_role:
                    ssh_cred_by_role[cred_role] = ssh_cred
                else:
                    errmsg = format_ssh_device_configuration_error(
                        "The SSH device had more than one credentials with the role '{}'".format(cred_role), sshdev_config)
                    raise AKitConfigurationError(errmsg) from None

            if "priv" not in ssh_cred_by_role:
                errmsg = format_ssh_device_configuration_error(
                        "All SSH devices must have a 'priv' credential.", sshdev_config)
                raise AKitConfigurationError(errmsg) from None

            priv_cred = ssh_cred_by_role["priv"]

            dev_type = sshdev_config["deviceType"]

            host = None
            upnp_hint = None
            if "host" in sshdev_config:
                host = sshdev_config["host"]
            elif dev_type == "network/upnp":
                upnp_config = sshdev_config["upnp"]

                if "hint" in upnp_config:
                    upnp_hint = upnp_config["hint"]
                elif "USN" in upnp_config:
                    upnp_hint = upnp_config["USN"]
                if upnp_coord is not None:
                    upnp_dev = upnp_coord.lookup_device_by_upnp_hint(upnp_hint)
                    dev = upnp_dev.basedevice
                    if dev is not None:
                        ipaddr = upnp_dev.IPAddress
                        host = ipaddr
                        sshdev_config["host"] = host
                        self._cl_upnp_hint_to_ip_lookup[upnp_hint] = ipaddr
                else:
                    ssh_config_errors.append(sshdev_config)

            if host is not None:
                called_id = dev.identity
                ip = socket.gethostbyname(host)
                self._cl_ip_to_host_lookup[ip] = host

                agent = SshAgent(host, priv_cred, users=ssh_cred_by_role, called_id=called_id)

                sshdev_config["ipaddr"] = agent.ipaddr
                try:
                    status, stdout, stderr = agent.run_cmd("echo Hello")
                    if status == 0 and stdout.strip() == "Hello":
                        ssh_devices_available.append(sshdev_config)
                    else:
                        ssh_devices_unavailable.append(sshdev_config)
                except:
                    ssh_devices_unavailable.append(sshdev_config)

                self._cl_children[host] = agent

                coord_ref = weakref.ref(self)

                basedevice = None
                if upnp_hint is not None:
                    basedevice = lscape._internal_lookup_device_by_hint(upnp_hint) # pylint: disable=protected-access
                else:
                    fdid = FriendlyIdentifier(host, host)
                    basedevice = lscape._create_landscape_device(fdid, dev_type, sshdev_config)
                    basedevice = lscape._enhance_landscape_device(basedevice, agent)
                    basedevice.initialize_features()

                basedevice.attach_extension("ssh", agent)

                basedevice_ref = weakref.ref(basedevice)
                agent.initialize(coord_ref, basedevice_ref, host, ip, sshdev_config)

                # If this device does not have a USN or hint then it is not a UPNP device so the
                # SshPoolCoordinator is repsonsible for activating it
                if upnp_hint is None:
                    lscape._internal_activate_device(host)

            else:
                ssh_config_errors.append(sshdev_config)

        self._available_devices = ssh_devices_available
        self._unavailable_devices = ssh_devices_unavailable

        return ssh_config_errors, ssh_devices_available, ssh_devices_unavailable

    def lookup_device_by_host(self, host: str) -> Union[LandscapeDevice, None]:
        """
            Looks up the agent for a device by its hostname.  If the
            agent is not found then the API returns None.

            :param host: The host name of the LandscapeDevice to search for.

            :returns: The found LandscapeDevice or None
        """
        device = None

        self._coord_lock.acquire()
        try:
            if host in self._cl_children:
                device = self._cl_children[host].basedevice
        finally:
            self._coord_lock.release()

        return device

    def lookup_device_by_ip(self, ip) -> Union[LandscapeDevice, None]:
        """
            Looks up the agent for a device by its ip address.  If the
            agent is not found then the API returns None.

            :param ip: The ip address of the LandscapeDevice to search for.

            :returns: The found LandscapeDevice or None
        """
        device = None

        self._coord_lock.acquire()
        try:
            if ip in self._cl_ip_to_host_lookup:
                if ip in self._cl_ip_to_host_lookup:
                    host = self._cl_ip_to_host_lookup[ip]
                    if host in self._cl_children:
                        device = self._cl_children[host].basedevice
        finally:
            self._coord_lock.release()

        return device

    def verify_connectivity(self, cmd: str = "echo 'It Works'", user: Optional[str] = None, raiseerror: bool = True) -> List[tuple]:
        """
            Loops through the nodes in the SSH pool and utilizes the credentials for the specified user in order to verify
            connectivity with the remote node.

            :param cmd: A command to run on the remote machine in order
                        to verify that ssh connectivity can be establish.
            :param user: The name of the user credentials to use for connectivity.
                         If the 'user' parameter is not provided, then the
                         credentials of the default or priviledged user will be used.
            :param raiseerror: A boolean value indicating if this API should raise an Exception on failure.

            :returns: A list of errors encountered when verifying connectivity with the devices managed or watched by the coordinator.
        """
        results = []

        for agent in self.children_as_extension:
            host = agent.host
            ipaddr = agent.ipaddr
            try:
                status, stdout, stderr = agent.run_cmd(cmd)
                results.append((host, ipaddr, status, stdout, stderr, None))
            except Exception as xcpt: # pylint: disable=broad-except
                if raiseerror:
                    raise
                results.append((host, ipaddr, None, None, None, xcpt))

        return results




__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import os
import sys

import click

from akit.cli.cmdtree.testing.constants import (
    HELP_ROOT,
    HELP_EXCLUDES,
    HELP_INCLUDES,
    HELP_DEBUG
)

@click.command("query", help="Command used to query information about tests.")
@click.option("--root", default=None, type=str, help=HELP_ROOT)
@click.option("--excludes", "-x", multiple=True, help=HELP_EXCLUDES)
@click.option("--includes", "-i", multiple=True, help=HELP_INCLUDES)
@click.option("--debug", default=False, type=bool, help=HELP_DEBUG)
def command_akit_testing_query(root, includes, excludes, debug):
    # pylint: disable=unused-import,import-outside-toplevel

    # We do the imports of the automation framework code inside the action functions because
    # we don't want to startup loggin and the processing of inputs and environment variables
    # until we have entered an action method.  Thats way we know how to setup the environment.

    # IMPORTANT: We need to load the context first because it will trigger the loading
    # of the default user configuration
    from akit.environment.context import Context
    from akit.environment.variables import extend_path, AKIT_VARIABLES

    ctx = Context()
    env = ctx.lookup("/environment")

    # Set the jobtype
    env["jobtype"] = "testrun"

    test_root = None
    if root is not None:
        AKIT_VARIABLES.AKIT_TESTROOT = root
    elif AKIT_VARIABLES.AKIT_TESTROOT is not None:
        root = AKIT_VARIABLES.AKIT_TESTROOT
    else:
        root = "."

    test_root = os.path.abspath(os.path.expandvars(os.path.expanduser(root)))
    if not os.path.isdir(test_root):
        errmsg = "The specified root folder does not exist. root=%s" % root
        if test_root != root:
            errmsg += " expanded=%s" % test_root
        raise click.BadParameter(errmsg)
    env["testroot"] = test_root

    # Make sure we extend PATH to include the test root
    extend_path(test_root)

    # We use console activation because all our input output is going through the terminal
    import akit.activation.console
    from akit.xlogging.foundations import logging_initialize, getAutomatonKitLogger

    # Initialize logging
    logging_initialize()
    logger = getAutomatonKitLogger()

    from akit.extensionpoints import AKitExtensionPoints
    akep = AKitExtensionPoints()

    # At this point in the code, we either lookup an existing test job or we create a test job
    # from the includes, excludes or test_module
    TestJobType = akep.get_testplus_default_job_type()
    result_code = 0
    with TestJobType(logger, test_root, includes=includes, excludes=excludes) as tjob:
        query_results = tjob.query()

        test_names = [tn for tn in query_results.keys()]
        test_names.sort()
        print()
        print("Tests:")
        for tname in test_names:
            tref = query_results[tname]
            print("    " + tname)
            param_names = [pn for pn in tref.subscriptions.keys()]
            param_names.sort()
            for pname in param_names:
                pinfo = tref.subscriptions[pname]
                print ("        {}: {}".format(pname, pinfo.describe_source()))


        print()

        if len(tjob.import_errors) > 0:
            print("IMPORT ERRORS:", file=sys.stderr)
            for ifilename in tjob.import_errors:
                imperr_msg = ifilename
                print("    " + imperr_msg, file=sys.stderr)
            print("", file=sys.stderr)

    return



"""
.. module:: eventedvariable
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the :class:`EventedVariable` class which is used to maintain
               variable subscription values and associated metadata.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2022, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Any, Tuple, Optional, TYPE_CHECKING

import time
import weakref

from datetime import datetime, timedelta

from akit.exceptions import AKitTimeoutError
from akit.interop.eventsinking.enumerations import EventedVariableState

EVENTVAR_WAIT_RETRY_INTERVAL = 1
EVENTVAR_WAIT_TIMEOUT = 60

if TYPE_CHECKING:
    from akit.interop.eventsinking.eventedvariablesink import EventedVariableSink

class EventedVariable:
    """
        The EventedVariable object is utilized to handle the storage and propagation
        of event values, updates and timeouts.

        NOTE: The properties associated with this object do not lock the
        subscription lock because they rely on eventual consistency.  They will
        always contain a value that was good at one point in time and will be
        updated in the future to a value that will be good for a new point in time.

        The "key" and "name" properties never change from thier initial values and
        they are set in the constructor when the object is under the control of only
        a single thread.

        If you need to ensure that the relationship between the value and updated
        members are in sync with each other, then a sync_read and sync_update
        API is provided to ensure this synchronization.
    """

    def __init__(self, key: str, name: str, sink_ref: weakref.ref, value: Any = None, data_type: Optional[str] = None, default: Any = None,
                 allowed_list=None, timestamp: datetime = None, evented: bool = True):
        """
            Constructor for the :class:`EventedVariable` object.

            :param key: The key {sink type}/{event name} for this event
            :param name: The name of the event this variable is storing information on.
            :param value: Optional initially reported value for the variable.  This is used when we have reports for
                          variables that we are not subscribed to.
            :param timestamp: The timestamp of the creation of this variable.  If a timestamp is passed then a value
                              needs to also be passed.
            :param evented: Indicates that this variable is evented to subscribers.
        """
        self._key = key
        self._name = name
        self._sink_ref = sink_ref
        self._value = value
        self._data_type = data_type
        self._default = default
        self._allowed_list = allowed_list
        self._timestamp = None
        
        if not evented:
            errmsg = "EventedVariable constructor was called for a variabled that is not evented."
            raise ValueError(errmsg)

        self._expires = None

        if self._value is not None and timestamp is None:
            self._timestamp  = datetime.now()

        if value is None and default is not None:
            self._value = default

        self._created = timestamp
        self._updated = timestamp
        self._changed = timestamp
        return

    @property
    def created(self) -> datetime:
        """
            When the event variabled value was set for the first time.
        """
        return self._created

    @property
    def changed(self) -> datetime:
        """
            A datetime object that indicates when the value was last changed in value.
        """
        return self._changed

    @property
    def expired(self) -> bool:
        """
            When the event variabled subscription has expired.
        """
        exp = False
        if self._expires is not None:
            now = datetime.now()
            if now > self._expires:
                exp = True
        else:
            exp = True

        return exp

    @property
    def key(self) -> str:
        """
            The key {sink type}/{event name} for this event.
        """
        return self._key

    @property
    def last_initiator_moment(self) -> Tuple[datetime, Any]:
        """
            The last initiator moment that was registered with the sink for this event.
        """
        moment = self.sink.initiator_moment_lookup(self._key)
        return moment

    @property
    def state(self) -> EventedVariableState:
        """
            The state of this event variable, UnInitialized, Valid or Stale
        """
        rtn_state = EventedVariableState.UnInitialized

        updated = self._updated
        if updated == datetime.min:
            rtn_state = EventedVariableState.Stale
        elif updated is not None:
            rtn_state = EventedVariableState.Valid

        return rtn_state

    @property
    def updated(self) -> datetime:
        """
            A datetime object that indicates when the value was last updated.
        """
        return self._updated

    @property
    def name(self) -> str:
        """
            The name of the event this variable is storing information on.
        """
        return self._name

    @property
    def sink(self) -> "EventedVariableSink":
        """
            Returns a reference to the owning :class:`EventedVariableSink` instance
        """
        return self._sink_ref()

    @property
    def value(self) -> Any:
        """
            The last value reported for the event variable this instance is referencing.
        """
        return self._value

    def invalidate_subscription(self):
        """
            Handles a invalidate subscription notification and sets the expiration field to
            now() to indicate that this EventedVariable is stale and will not receive
            any further updates.

            NOTE: After the invalidate has been received, the values of the variable
            can still be used, but should be used with the understanding that they are
            stale and should be used with caution.
        """
        self._expires = datetime.now()
        return

    def sync_read(self) -> Tuple[Any, datetime, datetime, EventedVariableState]:
        """
            Performs a threadsafe read of the value, updated, and state members of a
            :class:`EventedVariable` instance.

            ..note: 'sync_read' does not guarantee that the values being read are up-to-date.  If you want
            freshness guarantees and subscription assurance, you must use 'wait_for_update' or 'wait_for_value'
        """

        value, updated, changed, state = None, None, None, EventedVariableState.UnInitialized

        sink = self.sink
        for _ in sink.yield_state_lock():
            updated = self._updated
            changed = self._changed

            if updated == datetime.min:
                state = EventedVariableState.Stale
            elif updated is not None:
                state = EventedVariableState.Valid

            value = self._value

        return value, updated, changed, state

    def sync_update(self, value: Any, expires: Optional[datetime] = None, sink_locked: bool = False):
        """
            Peforms a threadsafe update of the value, updated and sid members of a
            :class:`EventedVariable` instance.
        """
        updated = datetime.now()

        if sink_locked:
            orig_value = self._value
            self._value, self._updated = value, updated
            if orig_value != self._value:
                self._changed = updated
            if expires is not None:
                self._expires = expires
        else:
            sink = self.sink
            for _ in sink.yield_state_lock():
                orig_value = self._value
                self._value, self._updated = value, updated
                if orig_value != self._value:
                    self._changed = updated
                if expires is not None:
                    self._expires = expires

        return

    def wait_for_update(self, moment: datetime, timeout: float = EVENTVAR_WAIT_TIMEOUT, interval: float = EVENTVAR_WAIT_RETRY_INTERVAL) -> Any:
        """
            Takes a datetime timestamp that is taken before a modification is made that
            will cause a state variable update and waits for the updated timestamp of
            this :class:`EventedVariable` instance to set to a timestamp that comes after the
            pre modification timestamp.

            :param moment: A timestamp taken from datetime.now() at a moment prior to engaging in
                            an activity that will result in a state variable change.
            :param timeout: The time in seconds to wait for the update to occur.
            :param interval: The time interval in seconds to wait before attempting to retry and
                             check to see if the updated timestamp has changed.

        """
        if self.expired:
            # If wait_for_update is being called, that means the caller wants a fresh
            # copy of the value for this event.  If this event is not evented, then
            # try renewing the sink subscription to see if we are given a value for
            # this variable when we renew our subscription.
            sink = self.sink
            if sink.auto_subscribe:
                sink.trigger_auto_subscribe_from_variable(self._key)

        now_time = datetime.now()
        start_time = now_time
        end_time = start_time + timedelta(seconds=timeout)
        while True:
            if now_time > end_time:
                raise AKitTimeoutError("Timeout waiting for event variable to update.") from None

            if self._updated is not None and self._updated > moment:
                break

            time.sleep(interval)
            now_time = datetime.now()

        return self._value

    def wait_for_value(self, timeout: float = EVENTVAR_WAIT_TIMEOUT, interval: float = EVENTVAR_WAIT_RETRY_INTERVAL) -> Any:
        """
            Waits for this :class:`EventedVariable` instance to contain a value.  It constains a
            value once the updated timestamp has been set.

            :param timeout: The time in seconds to wait for a value to be present.
            :param interval: The time interval in seconds to wait before attempting to retry and
                             check to see if the updated timestamp has been set.
        """
        if self.expired:
            # If wait_for_update is being called, that means the caller wants a fresh
            # copy of the value for this event.  If this event is not evented, then
            # try renewing the sink subscription to see if we are given a value for
            # this variable when we renew our subscription.
            sink = self.sink
            if sink.auto_subscribe:
                sink.trigger_auto_subscribe_from_variable(self._key)

        now_time = datetime.now()
        start_time = now_time
        end_time = start_time + timedelta(seconds=timeout)
        while True:
            if now_time > end_time:
                raise AKitTimeoutError("Timeout waiting for event variable to update.") from None

            if self._updated is not None:
                break
            time.sleep(interval)
            now_time = datetime.now()

        return self._value

    def __str__(self) -> str:
        value, updated, changed, state = self.sync_read()
        rtnstr = "name={} value={} updated={} changed={} state={}".format(self._name, value, updated, changed, state.name)
        return rtnstr



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import NamedTuple

from enum import IntEnum, Enum

REGION_CODES = [
    "AE", "AR", "AT", "AU", "BG", "BH", "BM", "BO", "BR", "CA", "CH", "CL",
    "CN", "CO", "CR", "CS", "CY", "CZ", "DE", "DK", "DO", "DZ", "EC", "EE",
    "EG", "ES", "FI", "FR", "GB", "GR", "GT", "HK", "HN", "ID", "IE", "IL",
    "IN", "IS", "IT", "JM", "JO", "KE", "KR", "KW", "KW", "LB", "LI", "LI",
    "LK", "LT", "LT", "LU", "MA", "MA", "MU", "MX", "MX", "NL", "NO", "NZ",
    "NZ", "OM", "PA", "PA", "PE", "PH", "PK", "PL", "PL", "PR", "PR", "PT",
    "QA", "RO", "RU", "RU", "SA", "SG", "SI", "SI", "SK", "SK", "SV", "TH",
    "TH", "TN", "TR", "TT", "TW", "UA", "US", "UY", "UY", "VE", "VN", "ZA"
]

DEFAULT_REGION_CODE = 'US'

class DeAuthenticationReason(IntEnum):
    """
        Reason code that is provided when an access point de-authenticates with a client. (IEEE 802.11)
    """
    PREV_AUTH_NOT_VALID = 2
    STA_LEAVE_IBBS = 3
    STA_INACTIVE = 4
    AP_OVERLOADED = 5
    STA_NOT_ASSOCIATED = 7
    STA_LEAVE_BSS = 8

class MfpMode(IntEnum):
    """
        OpenWRT AP settings for the Management Frame Protection (MFP) or
        Protected Management Frames (PMF) from standard IEEE 802.11w-2009.
    """
    DISABLED = 0
    OPTIONAL = 1
    REQUIRED = 2

class OpenWrtModel(str, Enum):
    NETGEAR_7600 = 'ipq806x'
    NETGEAR_WNDR3700 = 'ar71xx'
    NETGEAR_WNDR3700_1907 = 'ath79'


class WirelessEncryption(str, Enum):
    """
        Wireless encryption modes supported by routers running the OpenWRT firmware.
    """
    WEP = "wep"
    WEP_OPEN = "wep+open"
    WEP_MIXED = "wep+mixed"
    WPA2_TKIP_CCMP = "psk2+tkip+ccmp"
    WPA2_TKIP_AES = "psk2+tkip+aes"
    WPA2_TKIP = "psk2+tkip"
    WPA2_CCMP = "psk2+ccmp"
    WPA2_AES = "psk2+aes"
    WPA2 = "psk2"
    WPA_TKIP_CCMP = "psk+tkip+ccmp"
    WPA_TKIP_AES = "psk+tkip+aes"
    WPA_TKIP= "psk+tkip"
    WPA_CCMP = "psk+ccmp"
    WPA_AES = "psk+aes"
    WPA = "psk"
    WPA_MIXED_TKIP_CCMP = "mixed-psk+tkip+ccmp"
    WPA_MIXED_TKIP_AES = "mixed-psk+tkip+aes"
    WPA_MIXED = "mixed-psk"
    WPA_MIXED_TKIP = "mixed-psk+tkip"
    WPA_MIXED_CCMP = "mixed-psk+ccmp"
    WPA_MIXED_AES = "mixed-psk+aes"
    WEP_SHARED = "wep-shared"
    OPEN = "none"
    WPA3 = "sae"
    WPA3_MIXED = "sae-mixed"



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class HVAC_UserOperatingMode1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:HVAC_UserOperatingMode:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'HVAC_UserOperatingMode1'

    SERVICE_DEFAULT_VARIABLES = {
        "ModeStatus": { "data_type": "string", "default": "Off", "allowed_list": "['Off', 'InDeadBand', 'HeatOn', 'CoolOn', 'AutoChangeOver', 'AuxHeatOn', 'EconomyHeatOn', 'EmergencyHeatOn', 'AuxCoolOn', 'EconomyCoolOn', 'BuildingProtection', 'EnergySavingsHeating', 'EnergySavingsCooling']"},
        "ModeTarget": { "data_type": "string", "default": "Off", "allowed_list": "['Off', 'HeatOn', 'CoolOn', 'AutoChangeOver', 'AuxHeatOn', 'EconomyHeatOn', 'EmergencyHeatOn', 'AuxCoolOn', 'EconomyCoolOn', 'BuildingProtection', 'EnergySavingsMode']"},
        "Name": { "data_type": "string", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {}

    def action_GetModeStatus(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetModeStatus action.

            :returns: "CurrentModeStatus"
        """
        arguments = { }

        out_params = self.call_action("GetModeStatus", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentModeStatus",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetModeTarget(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetModeTarget action.

            :returns: "CurrentModeTarget"
        """
        arguments = { }

        out_params = self.call_action("GetModeTarget", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentModeTarget",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetName(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetName action.

            :returns: "CurrentName"
        """
        arguments = { }

        out_params = self.call_action("GetName", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentName",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetModeTarget(self, NewModeTarget, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetModeTarget action.
        """
        arguments = {
            "NewModeTarget": NewModeTarget,
        }

        self.call_action("SetModeTarget", arguments=arguments, aspects=aspects)

        return

    def action_SetName(self, NewName, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetName action.
        """
        arguments = {
            "NewName": NewName,
        }

        self.call_action("SetName", arguments=arguments, aspects=aspects)

        return



"""
.. module:: testrun
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module that is utilized by test runs to activate the test environment
               for use by a test run.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import os

from akit.exceptions import AKitSemanticError
from akit.environment.variables import ActivationProfile, AKIT_VARIABLES, JOB_TYPES

__activation_profile__ = ActivationProfile.TestRun

# Guard against attemps to activate more than one, activation profile.
if AKIT_VARIABLES.AKIT_ACTIVATION_PROFILE is not None:
    errmsg = "An attempt was made to activate multiple environment activation profiles. profile={}".format(
        AKIT_VARIABLES.AKIT_ACTIVATION_PROFILE
    )
    raise AKitSemanticError(errmsg)

AKIT_VARIABLES.AKIT_ACTIVATION_PROFILE = ActivationProfile.TestRun

AKIT_VARIABLES.AKIT_JOBTYPE = JOB_TYPES.TESTRUN

os.environ["AKIT_JOBTYPE"] = AKIT_VARIABLES.AKIT_JOBTYPE

import akit.activation.base



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Callable

import threading
import weakref

class TrafficCaptureContext:

    def __init__(self, owner: object, identifier: str, fromip: str, destroy_method: Callable):
        self._owner_ref = weakref.ref(owner)
        self._identifier = identifier
        self._fromip = fromip
        self._destroy_method = destroy_method
        self._lock = threading.Lock()
        self._captures = []
        return

    def __enter__(self):
        return self
    
    def __exit__(self, ex_type, ex_inst, ex_tb):
        self._destroy_method(self)
        return False

    @property
    def captures(self):
        rtnlist = None

        self._lock.acquire()
        try:
            rtnlist = [c for c in self._captures]
        finally:
            self._lock.release()

        return rtnlist
    
    @property
    def owner(self):
        return self._owner_ref()

    @property
    def fromip(self):
        self._fromip

    @property
    def identifier(self):
        return self._identifier

    def append_capture(self, req_headers: dict, req_body: str):

        self._lock.acquire()
        try:
            self._captures.append((req_headers, req_body))
        finally:
            self._lock.release()

        return



__author__ = "Myron Walker"
__copyright__ = "Copyright 2022, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import click

from akit.cli.cmdtree.servers.results_server import command_akit_servers_results_server

@click.group("servers", help="Contains command for working with tests and test results.")
def group_akit_servers():
    return

group_akit_servers.add_command(command_akit_servers_results_server)



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Optional, Tuple

import socket

from socketserver import BaseServer
from threading import Thread

from http.server import HTTPServer, BaseHTTPRequestHandler

from akit.xthreading.looper import Looper
from akit.xthreading.looperpool import LooperPool

class SimpleHTTPLooper(Looper):
    """
    """

    def loop(self, packet) -> bool:
        """
            Method that is overloaded by derived classes in order to implement a work loop.
        """
        server, request, client_address = packet
        self.process_request_packet(server, request, client_address)
        return True

    def process_request_packet(self, server, request, client_address):
        """
            Same as in BaseServer but in a looper thread.  We also perform some
            exception handling here to prevent threads from shutting down unexpectedly.

        """
        orig_name = self.thread_get_name()
        self.thread_set_name("{}-*".format(orig_name))
        try:
            server.finish_request(request, client_address)
        except Exception:
            server.handle_error(request, client_address)
        finally:
            server.shutdown_request(request)
        self.thread_set_name(orig_name)
        return

class HttpRequestHandler(BaseHTTPRequestHandler):
    def __init__(self, request: socket.socket, client_address: Tuple[str, int], server: BaseServer, **_kwargs) -> None:
        """
            ..note: Overide the constructor for BaseHTTPRequestHandler so we can absorb any extra kwargs.
        """
        BaseHTTPRequestHandler.__init__(request, client_address, server)
        self._kwargs = _kwargs
        return

class HttpServerThreadPool(LooperPool, HTTPServer):
    def __init__(self, address: Tuple[str, int], handler_class: HttpRequestHandler, group_name: str='webserver-worker',
                 min_loopers: int=5, max_loopers: int=10, highwater: int=5, daemon=True, **handler_kwargs):
        LooperPool.__init__(self, SimpleHTTPLooper, group_name=group_name, min_loopers=min_loopers, max_loopers=max_loopers,
                            highwater=highwater, daemon=daemon)
        HTTPServer.__init__(self, address, handler_class)
        self._handler_kwargs = handler_kwargs
        return

    def finish_request(self, request: socket.socket, client_address: Tuple[str, int]) -> None:
        self.RequestHandlerClass(request, client_address, self, **self._handler_kwargs)
        return

    def get_request(self) -> socket.socket:
        asock = self.socket.accept()
        return asock

    def get_server_address(self) -> Tuple[str, int]:
        rtnval = self.server_address
        return rtnval

    def process_request(self, request, client_address):
        """
            Start a new thread to process the request.
        """
        self.push_work((self, request, client_address))
        return

    def server_stop(self):
        super().server_close()
        self.shutdown()
        return

    def server_start(self):
        self.start_pool()

        start_thread = Thread(target=self.serve_forever, name=self._group_name)
        start_thread.daemon = True
        start_thread.start()
        return
